import time

from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.platforms.lisy import lisy

from mpf.tests.MpfTestCase import MpfTestCase, test_config, MagicMock
from mpf.tests.loop import MockSerial, MockSocket

COMMAND_LENGTH = {
    0: 1,
    1: 1,
    2: 1,
    3: 1,
    4: 1,
    6: 1,
    7: 2,
    9: 1,
    11: 2,
    12: 2,
    19: 1,
    21: 2,
    22: 2,
    23: 2,
    24: 3,
    25: 3,
    40: 2,
    41: 1,
    51: 1,
    60: 11,
    100: 1,
    101: 1,
}


class MockLisySocket(MockSocket, MockSerial):

    def read(self, length):
        del length
        if not self.queue:
            return b""
        msg = self.queue.pop()
        return msg

    def read_ready(self):
        return bool(self.queue)

    def write_ready(self):
        return True

    def write(self, msg):
        """Write message."""
        # print("Serial received: " + "".join("\\x%02x" % b for b in msg) + " len: " + str(len(msg)))
        total_msg_len = len(msg)
        while msg:
            command = msg[0]
            if command == 13:
                command_length = 6 + msg[5]
            elif 30 <= command <= 35:
                if self.api_version >= 9:
                    command_length = 2 + msg[1]
                else:
                    command_length = msg.index(b'\x00') + 1
            elif command in (50, 54):
                command_length = 3 if self.api_version >= 9 else 2
            elif command == 51:
                command_length = 2 if self.api_version >= 9 else 1
            elif command in (52, 53):
                command_length = msg[3:].index(b'\x00') + 4
            else:
                command_length = COMMAND_LENGTH.get(command, None)
            if command_length is None:
                raise AssertionError("Unknown command {} in {}".format(command, "".join("\\x%02x" % b for b in msg)))
            if len(msg) < command_length:
                raise AssertionError("Message too short ({}) for command {} with length {} in {}".format(
                    len(msg), command_length, command,
                    "".join("\\x%02x" % b for b in msg)))

            self._handle_msg(msg[0:command_length])
            msg = msg[command_length:]

        return total_msg_len

    def _handle_msg(self, msg):
        if msg in self.permanent_commands and msg not in self.expected_commands:
            if self.permanent_commands[msg] is not None:
                self.queue.append(self.permanent_commands[msg])
            return len(msg)

        if msg not in self.expected_commands:
            self.crashed = True
            print("Unexpected command: " + "".join("\\x%02x" % b for b in msg) + " len: " + str(len(msg)))
            raise AssertionError("Unexpected command: " + "".join("\\x%02x" % b for b in msg) +
                                 " len: " + str(len(msg)))

        if self.expected_commands[msg] is not None:
            self.queue.append(self.expected_commands[msg])

        del self.expected_commands[msg]
        return len(msg)

    def send(self, data):
        return self.write(data)

    def recv(self, size):
        return self.read(size)

    def __init__(self, api_version):
        super().__init__()
        self.name = "SerialMock"
        self.expected_commands = {}
        self.queue = []
        self.permanent_commands = {}
        self.crashed = False
        self.api_version = api_version


class TestLisy(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/lisy/'

    def _mock_loop(self):
        self.clock.mock_socket("localhost", 1234, self.serialMock)

    def tearDown(self):
        self.assertFalse(self.serialMock.crashed)
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while self.serialMock.expected_commands and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)

    def setUp(self):
        self.expected_duration = 1.5
        self.serialMock = MockLisySocket(api_version=8)

        self.serialMock.permanent_commands = {
            b'\x29': b'\x7F',           # changed switches? -> no
            b'\x65': b'\x00'            # watchdog
        }

        self.serialMock.expected_commands = {
            b'\x00': b'LISY1\00',       # hw LISY1
            b'\x01': b'4.01\00',        # lisy version
            b'\x02': b'0.08\00',        # api version
            b'\x64': b'\x00',           # reset -> ok
            b'\x03': b'\x28',           # get number of lamps -> 40
            b'\x04': b'\x09',           # get number of solenoids -> 9
            b'\x06': b'\x05',           # get number of displays -> 5
            b'\x09': b'\x58',           # get number of switches -> 88
            b'\x1e\x00': None,          # clear display
            b'\x1f\x00': None,          # clear display
            b'\x20\x00': None,          # clear display
        }

        for number in range(88):
            if number % 10 >= 8:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x02'
            else:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x00' if number != 37 else b'\x01'

        super().setUp()

        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def test_platform(self):
        self.maxDiff = None
        infos = """LISY connected via network at localhost:1234
Hardware: LISY1 Lisy Version: 4.01 API Version: 0.8
Input count: 88 Input map: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37', '38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49', '50', '51', '52', '53', '54', '55', '56', '57', '58', '59', '60', '61', '62', '63', '64', '65', '66', '67', '68', '69', '70', '71', '72', '73', '74', '75', '76', '77', '78', '79', '80', '81', '82', '83', '84', '85', '86', '87']
Coil count: 9
Modern lights count: 0
Traditional lights count: 40
Display count: 5
"""

        self.assertEqual(self.machine.default_platform.get_info_string(), infos)

        # wait for watchdog
        self.serialMock.expected_commands = {
            b'\x65': b'\x00'            # watchdog
        }
        self._wait_for_processing()

        # test initial switch state
        self.assertSwitchState("s_test00", False)
        self.assertSwitchState("s_test37", True)
        self.assertSwitchState("s_test77_nc", True)

        self.serialMock.expected_commands = {
            b'\x29': b'\x25'        # 37 turned inactive
        }
        self.advance_time_and_run(.1)
        # turns inactive
        self.assertSwitchState("s_test37", False)

        self.serialMock.expected_commands = {
            b'\x29': b'\xA5'        # 37 turned active (again)
        }
        self.advance_time_and_run(.1)
        # turns active
        self.assertSwitchState("s_test37", True)

        self.serialMock.expected_commands = {
            b'\x29': b'\xCD'        # 77 turned active
        }
        self.advance_time_and_run(.1)
        # turns inactive (because of NC)
        self.assertSwitchState("s_test77_nc", False)

        # pulse coil
        self.serialMock.expected_commands = {
            b'\x18\x00\x0a': None,      # set pulse_ms to 10ms
            b'\x17\x00': None
        }
        self.machine.coils["c_test"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # pulse trough eject. enable and disable in software
        self.serialMock.expected_commands = {
            b'\x18\x67\x00': None,  # set pulse_ms to 10ms
            b'\x15\x67': None       # enable
        }
        self.machine.coils["c_trough_eject"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.advance_time_and_run(2)
        self.serialMock.expected_commands = {
            b'\x16\x67': None       # disable
        }
        self.advance_time_and_run()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # enable coil
        self.serialMock.expected_commands = {
            b'\x18\x01\x0a': None,  # set pulse_ms to 10ms
            b'\x15\x01': None
        }
        self.machine.coils["c_test_allow_enable"].enable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # disable coil
        self.serialMock.expected_commands = {
            b'\x16\x01': None
        }
        self.machine.coils["c_test_allow_enable"].disable()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test light enable (using light 3)
        self.serialMock.expected_commands = {
            b'\x0b\x03': None
        }
        self.machine.lights["test_light"].on(key="test")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # disable light (using light 3)
        self.serialMock.expected_commands = {
            b'\x0c\x03': None
        }
        self.machine.lights["test_light"].remove_from_stack_by_key("test")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # start ball. enable flipper (using light 1)
        self.serialMock.expected_commands = {
            b'\x0b\x01': None
        }
        self.post_event("ball_started")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.advance_time_and_run()

        # end ball. disable flipper (using light 1)
        self.serialMock.expected_commands = {
            b'\x0c\x01': None
        }
        self.post_event("ball_will_end")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set info display to TEST
        self.serialMock.expected_commands = {
            b'\x1ETEST\x00': None
        }
        self.machine.segment_displays["info_display"].add_text("TEST")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set player 1 display to 42000
        self.serialMock.expected_commands = {
            b'\x1F42000\x00': None
        }
        self.machine.segment_displays["player1_display"].add_text("42000")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set player 1 display to flashing
        self.serialMock.expected_commands = {
            b'\x1F42000\x00': None,
        }
        self.machine.segment_displays["player1_display"].set_flashing(FlashingType.FLASH_ALL)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x1F42000\x00': None,
            b'\x1F\x00': None
        }

        self.advance_time_and_run(1)
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x1F42000\x00': None,
        }
        self.machine.segment_displays["player1_display"].set_flashing(FlashingType.NO_FLASH)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound
        self.serialMock.expected_commands = {
            b'\x32\x02': None
        }
        self.post_event("test2")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound file
        self.serialMock.expected_commands = {
            b'\x34\x00some_file\x00': None
        }
        self.post_event("play_file")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound file looping
        self.serialMock.expected_commands = {
            b'\x34\x01some_file\x00': None
        }
        self.post_event("play_file_loop")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # text to speech
        self.serialMock.expected_commands = {
            b'\x35\x02Hello MPF\x00': None
        }
        self.post_event("play_text")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set volume to 50 (32 hex)
        self.serialMock.expected_commands = {
            b'\x36\x32': None
        }
        self.post_event("volume_05")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # increase volume by 0.1 -> 60 -> hex 3C
        self.serialMock.expected_commands = {
            b'\x36\x3C': None
        }
        self.post_event("increase_volume")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # decrease volume by 0.01 -> 59 -> hex 3B
        self.serialMock.expected_commands = {
            b'\x36\x3B': None
        }
        self.post_event("decrease_volume")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test another sound
        self.serialMock.expected_commands = {
            b'\x32\x03': None
        }
        self.post_event("test3")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # stop sound
        self.serialMock.expected_commands = {
            b'\x33': None
        }
        self.post_event("test_stop")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    @test_config("config_system11.yaml")
    def test_system11(self):
        # test normal coil
        self.serialMock.expected_commands = {
            b'\x18\x00\x14': None,  # set pulse_ms to 20ms
            b'\x17\x00': None       # pulse coil 0
        }
        self.machine.coils["c_test"].pulse(20)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test C-side coil
        self.serialMock.expected_commands = {
            b'\x18\x08\x0a': None,  # set pulse_ms to 10ms to A/C relay
            b'\x15\x08': None,      # enable A/C relay
            b'\x18\x01\x14': None,  # set pulse_ms to 20ms
            b'\x17\x01': None       # pulse coil 1
        }
        self.machine.coils["c_test1_c_side"].pulse(20)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # wait for A/C disable
        self.serialMock.expected_commands = {
            b'\x16\x08': None,      # disable A/C relay
        }
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test C-side coil
        self.serialMock.expected_commands = {
            b'\x15\x08': None,      # enable A/C relay
            b'\x17\x01': None       # pulse coil 1
        }
        self.machine.coils["c_test1_c_side"].pulse(20)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # wait for A/C disable
        self.serialMock.expected_commands = {
            b'\x16\x08': None,      # disable A/C relay
        }
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test A-side coil
        self.serialMock.expected_commands = {
            b'\x17\x01': None       # pulse coil 1
        }
        self.machine.coils["c_test1_a_side"].pulse(20)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test C-side coil
        self.serialMock.expected_commands = {
            b'\x18\x01\x0f': None,  # set pulse_ms to 15ms
            b'\x17\x01': None       # pulse coil 1
        }
        self.machine.coils["c_test1_a_side"].pulse(15)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test C-side coil
        self.serialMock.expected_commands = {
            b'\x15\x08': None,      # enable A/C relay
            b'\x18\x01\x14': None,  # set pulse_ms to 20ms
            b'\x17\x01': None       # pulse coil 1
        }
        self.machine.coils["c_test1_c_side"].pulse(20)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # wait for A/C disable
        self.serialMock.expected_commands = {
            b'\x16\x08': None,      # disable A/C relay
        }
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)


class TestAPC(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/apc/'

    def _mock_loop(self):
        self.clock.mock_serial("com1", self.serialMock)

    def tearDown(self):
        self.assertFalse(self.serialMock.crashed)
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while self.serialMock.expected_commands and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)

    def setUp(self):
        self.expected_duration = 1.5
        self.serialMock = MockLisySocket(api_version=9)

        self.serialMock.permanent_commands = {
            b'\x29': b'\x7F',           # changed switches? -> no
            b'\x65': b'\x00'            # watchdog
        }

        self.serialMock.expected_commands = {
            b'\x00': b'APC\00',         # hw APC
            b'\x01': b'0.02\00',        # APC version
            b'\x02': b'0.09\00',        # api version
            b'\x64': b'\x00',           # reset -> ok
            b'\x03': b'\x28',           # get number of lamps -> 40
            b'\x04': b'\x09',           # get number of solenoids -> 9
            b'\x06': b'\x05',           # get number of displays -> 5
            b'\x07\x00': b'\x02\x10',   # get type of display 0
            b'\x07\x01': b'\x03\x05',   # get type of display 1
            b'\x07\x02': b'\x04\x07',   # get type of display 2
            b'\x07\x03': b'\x05\x03',   # get type of display 3
            b'\x07\x04': b'\x06\x10',   # get type of display 4
            b'\x09': b'\x58',           # get number of switches -> 88
            b'\x13': b'\x00',           # get number of modern lights -> 0
            b'\x1e\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,  # clear display
            b'\x1f\x05\x00\x00\x00\x00\x00': None,                                              # clear display
            b'\x20\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,          # clear display
            b'\x21\x03\x20\x20\x20': None,                                                      # clear display
            b'\x22\x10\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20': None,  # clear display
            b'\x19\x00\x0a': None,
            b'\x19\x01\x0a': None,
            b'\x19\x67\xff': None,
            b'\x19\x05\x1e': None,
            b'\x19\x06\x0a': None,
            b'\x19\x07\x0a': None,
        }

        for number in range(88):
            if number % 10 >= 8:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x02'
            else:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x00' if number != 37 else b'\x01'

        # prevent changes on real hardware
        lisy.LisyHardwarePlatform._disable_dts_on_start_of_serial = MagicMock()

        super().setUp()

        lisy.LisyHardwarePlatform._disable_dts_on_start_of_serial.assert_called_with()

        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def test_platform(self):
        # wait for watchdog
        self.serialMock.expected_commands = {
            b'\x65': b'\x00'            # watchdog
        }
        self._wait_for_processing()

        # test sound
        self.serialMock.expected_commands = {
            b'\x32\x01\x02': None
        }
        self.post_event("test2")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound on track 2
        self.serialMock.expected_commands = {
            b'\x32\x02\x05': None
        }
        self.post_event("test4")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound file
        self.serialMock.expected_commands = {
            b'\x34\x01\x00some_file\x00': None
        }
        self.post_event("play_file")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test sound file looping
        self.serialMock.expected_commands = {
            b'\x34\x01\x01some_file\x00': None
        }
        self.post_event("play_file_loop")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # text to speech
        self.serialMock.expected_commands = {
            b'\x35\x01\x02Hello MPF\x00': None
        }
        self.post_event("play_text")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set volume to 50 (32 hex)
        self.serialMock.expected_commands = {
            b'\x36\x01\x32': None
        }
        self.post_event("volume_05")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # increase volume by 0.1 -> 60 -> hex 3C
        self.serialMock.expected_commands = {
            b'\x36\x01\x3C': None
        }
        self.post_event("increase_volume")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # decrease volume by 0.01 -> 59 -> hex 3B
        self.serialMock.expected_commands = {
            b'\x36\x01\x3B': None
        }
        self.post_event("decrease_volume")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test another sound
        self.serialMock.expected_commands = {
            b'\x32\x01\x03': None
        }
        self.post_event("test3")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # stop sound
        self.serialMock.expected_commands = {
            b'\x33\01': None
        }
        self.post_event("test_stop")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def test_rules(self):
        """Test HW Rules."""
        self.serialMock.expected_commands = {
            b'\x3c\x05\x01\x02\x00\x1e\xff\x00\x03\x02\x00': None,      # create rule for main
            b'\x3c\x06\x01\x00\x00\x0a\xff\xff\x03\x00\x00': None,      # create rule for hold
        }
        self.machine.flippers["f_test_hold_eos"].enable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for main
            b'\x3c\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for hold
        }
        self.machine.flippers["f_test_hold_eos"].disable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x07\x03\x00\x00\x0a\xff\x00\x01\x00\x00': None,      # add rule for slingshot
            b'\x19\x07\x14': None
        }
        self.machine.autofires["ac_slingshot"].enable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for slingshot
        }
        self.machine.autofires["ac_slingshot"].disable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set info display to TEST
        self.serialMock.expected_commands = {
            b'\x1e\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x03\x03\x07': None
        }
        self.machine.segment_displays["info_display"].add_text("1337")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test recycle
        # pulse coil
        self.serialMock.expected_commands = {
            b'\x18\x00\x0a': None,      # set pulse_ms to 10ms
            b'\x17\x00': None
        }
        self.machine.coils["c_test"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)


class TestLisyV10(MpfTestCase):

    def get_config_file(self):
        return 'config_modern.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/lisy/'

    def _mock_loop(self):
        self.clock.mock_serial("com1", self.serialMock)

    def tearDown(self):
        self.assertFalse(self.serialMock.crashed)
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while self.serialMock.expected_commands and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)

    def setUp(self):
        self.expected_duration = 1.5
        self.serialMock = MockLisySocket(api_version=10)

        self.serialMock.permanent_commands = {
            b'\x29': b'\x7F',           # changed switches? -> no
            b'\x65': b'\x00'            # watchdog
        }

        self.serialMock.expected_commands = {
            b'\x00': b'FUTURE_HARDWARE\00',         # hw not known yet
            b'\x01': b'0.42\00',        # hardware version
            b'\x02': b'0.10\00',        # api version
            b'\x64': b'\x00',           # reset -> ok
            b'\x03': b'\x00',           # get number of simple lamps -> 0
            b'\x04': b'\x09',           # get number of solenoids -> 9
            b'\x06': b'\x00',           # get number of displays -> 0
            b'\x09': b'\x58',           # get number of switches -> 88
            b'\x13': b'\x02\x01',       # get number of modern lights -> 512 + 1 = 513 modern lights
            b'\x19\x00\x0a': None,
            b'\x19\x01\x0a': None,
            b'\x19\x05\x1e': None,
            b'\x19\x06\x0a': None,
            b'\x19\x07\x0a': None,
        }

        for number in range(88):
            if number % 10 >= 8:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x02'
            else:
                self.serialMock.expected_commands[bytes([40, number])] = b'\x00' if number != 37 else b'\x01'

        # prevent changes on real hardware
        lisy.LisyHardwarePlatform._disable_dts_on_start_of_serial = MagicMock()

        super().setUp()

        lisy.LisyHardwarePlatform._disable_dts_on_start_of_serial.assert_called_with()

        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def test_rules(self):
        """Test HW Rules."""
        # wait for watchdog
        self.serialMock.expected_commands = {
            b'\x65': b'\x00'  # watchdog
        }
        self._wait_for_processing()

        self.serialMock.expected_commands = {
            b'\x3c\x05\x01\x02\x00\x1e\xff\x00\x03\x02\x00': None,      # create rule for main
            b'\x3c\x06\x01\x00\x00\x0a\xff\xff\x03\x00\x00': None,      # create rule for hold
        }
        self.machine.flippers["f_test_hold_eos"].enable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for main
            b'\x3c\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for hold
        }
        self.machine.flippers["f_test_hold_eos"].disable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x07\x03\x00\x00\x0a\xff\x00\x01\x00\x00': None,      # add rule for slingshot
            b'\x19\x07\x14': None
        }
        self.machine.autofires["ac_slingshot"].enable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'\x3c\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00': None,      # remove rule for slingshot
        }
        self.machine.autofires["ac_slingshot"].disable()
        self.advance_time_and_run(.2)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # test recycle
        # pulse coil
        self.serialMock.expected_commands = {
            b'\x18\x00\x0a': None,      # set pulse_ms to 10ms
            b'\x17\x00': None
        }
        self.machine.coils["c_test"].pulse()
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

    def test_lights(self):
        """Test lights."""
        # set color to one light without fade
        self.serialMock.expected_commands = {
            b'\x0d\x00\x00\x00\x00\x03\x11\x22\x33': None,      # fade with 0ms fade time
        }
        self.machine.lights["test_light0"].color([0x11, 0x22, 0x33], key="test")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set color again (should do nothing)
        self.machine.lights["test_light0"].remove_from_stack_by_key("test")
        self.machine.lights["test_light0"].color([0x11, 0x22, 0x33], fade_ms=100)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set color to the second light without fade
        self.serialMock.expected_commands = {
            b'\x0d\x00\x03\x00\x00\x04\x11\x22\x33\x11': None,      # fade with 0ms fade time starting at channel 3
                                                                    # 4 channels because this is a RGBW light
        }
        self.machine.lights["test_light1"].color([0x11, 0x22, 0x33])
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # fade both lights together (fade depending on serial timing)
        self.serialMock.expected_commands = {
            b'\x0d\x00\x00\x01\x18\x07\xaa\xbb\xcc\xdd\xee\xff\xdd': None,      # fade with 300ms fade time
            b'\x0d\x00\x00\x01\x19\x07\xaa\xbb\xcc\xdd\xee\xff\xdd': None,      # fade with 300ms fade time
            b'\x0d\x00\x00\x01\x20\x07\xaa\xbb\xcc\xdd\xee\xff\xdd': None,      # fade with 300ms fade time
            b'\x0d\x00\x00\x01\x21\x07\xaa\xbb\xcc\xdd\xee\xff\xdd': None,      # fade with 300ms fade time
            b'\x0d\x00\x00\x01\x22\x07\xaa\xbb\xcc\xdd\xee\xff\xdd': None,      # fade with 300ms fade time
        }
        self.machine.lights["test_light0"].color([0xaa, 0xbb, 0xcc], fade_ms=300)
        self.machine.lights["test_light1"].color([0xdd, 0xee, 0xff], fade_ms=300)
        start = time.time()
        while len(self.serialMock.expected_commands) > 4 and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)
        self.assertEqual(4, len(self.serialMock.expected_commands))



"""Classes for the EventManager and QueuedEvents."""
import inspect
from collections import deque, namedtuple, defaultdict
import uuid

import asyncio
from functools import partial, lru_cache
from unittest.mock import MagicMock

from typing import Dict, Any, Tuple, Optional, Callable, List

from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.placeholder_manager import BaseTemplate   # pylint: disable-msg=cyclic-import,unused-import
    from typing import Deque    # pylint: disable-msg=cyclic-import,unused-import

EventHandlerKey = namedtuple("EventHandlerKey", ["key", "event"])
RegisteredHandler = namedtuple("RegisteredHandler", ["callback", "priority", "kwargs", "key", "condition",
                                                     "blocking_facility"])
PostedEvent = namedtuple("PostedEvent", ["event", "type", "callback", "kwargs"])


class EventManager(MpfController):

    """Handles all the events and manages the handlers in MPF."""

    config_name = "event_manager"

    __slots__ = ["registered_handlers", "event_queue", "callback_queue", "monitor_events", "_queue_tasks"]

    def __init__(self, machine: "MachineController") -> None:
        """Initialize EventManager."""
        super().__init__(machine)

        self.registered_handlers = defaultdict(list)    # type: Dict[str, List[RegisteredHandler]]
        self.event_queue = deque([])        # type: Deque[PostedEvent]
        self.callback_queue = deque([])     # type: Deque[Tuple[Any, dict]]
        self.monitor_events = False
        self._queue_tasks = []              # type: List[asyncio.Task]

        self.add_handler("debug_dump_stats", self._debug_dump_events)

    def _debug_dump_events(self, **kwargs):
        del kwargs
        self.log.info("--- DEBUG DUMP EVENTS ---")
        self.log.info("Total registered_handlers: %s. Total event_queue: %s. Total callback_queue: %s. "
                      "Total _queue_tasks: %s", len(self.registered_handlers), len(self.event_queue),
                      len(self.callback_queue), len(self._queue_tasks))
        self.log.info("Registered Handlers:")
        handlers = sorted(self.registered_handlers.items(), key=lambda x: -len(x[1]))
        for event_name, event_list in handlers:
            self.log.info("  Total handlers: %s (for %s)", len(event_list), event_name)

        self.log.info("Queue events:")
        for event_task in self._queue_tasks:
            self.log.info(" %s:", event_task)

        self.log.info("--- DEBUG DUMP EVENTS END ---")

    @lru_cache()
    def get_event_and_condition_from_string(self, event_string: str) -> Tuple[str, Optional["BaseTemplate"], int]:
        """Parse an event string to divide the event name from a possible placeholder / conditional in braces.

        Args:
        ----
            event_string: String to parse

        Returns 2-item tuple- First item is the event name. Second item is the
        condition (A BoolTemplate instance) if it exists, or None if it doesn't.
        """
        placeholder = None
        additional_priority = 0
        if event_string[-1:] == "}":
            first_bracket_pos = event_string.find("{")
            if first_bracket_pos < 0:
                raise ValueError('Failed to parse condition in event name, '
                                 'please remedy "{}"'.format(event_string))
            if " " in event_string[0:first_bracket_pos]:
                raise ValueError('Cannot handle events with spaces in the event name, '
                                 'please remedy "{}"'.format(event_string))
            placeholder = self.machine.placeholder_manager.build_bool_template(event_string[first_bracket_pos + 1:-1])
            event_string = event_string[0:first_bracket_pos]
        else:
            if " " in event_string:
                raise ValueError('Cannot handle events with spaces in the event name, '
                                 'please remedy "{}"'.format(event_string))
            if "{" in event_string:
                raise ValueError('Failed to parse condition in event name, '
                                 'please remedy "{}"'.format(event_string))

        priority_start = event_string.find(".")
        if priority_start > 0:
            additional_priority = int(event_string[priority_start + 1:])
            event_string = event_string[:priority_start]

        return event_string, placeholder, additional_priority

    def add_async_handler(self, event: str, handler: Any, priority: int = 1, blocking_facility: Any = None,
                          **kwargs) -> EventHandlerKey:
        """Register a coroutine as event handler."""
        return self.add_handler(event, partial(self._async_handler_coroutine, handler), priority, blocking_facility,
                                **kwargs)

    def _async_handler_coroutine(self, _coroutine, queue, **kwargs):
        queue.wait()
        task = self.machine.clock.loop.create_task(_coroutine(**kwargs))
        task.add_done_callback(partial(self._async_handler_done, queue))

    @staticmethod
    def _async_handler_done(queue, future):
        try:
            future.result()
        except asyncio.CancelledError:
            pass
        queue.clear()

    def add_handler(self, event: str, handler: Any, priority: int = 1, blocking_facility: Any = None,
                    **kwargs) -> EventHandlerKey:
        """Register an event handler to respond to an event.

        Args:
        ----
            event: String name of the event you're adding a handler for. Since
                events are text strings, they don't have to be pre-defined.
            handler: The callable method that will be called when the event is
                fired. Since it's possible for events to have kwargs attached
                to them, the handler method must include ``**kwargs`` in its
                signature.
            priority: An arbitrary integer value that defines what order the
                handlers will be called in. The default is 1, so if you have a
                handler that you want to be called first, add it here with a
                priority of 2. (Or 3 or 10 or 100000.) The numbers don't matter.
                They're called from highest to lowest. (i.e. priority 100 is
                called before priority 1.)
            blocking_facility: Facility which can block this event.
            **kwargs: Any any additional keyword/argument pairs entered here
                will be attached to the handler and called whenever that
                handler is called. Note these are in addition to kwargs that
                could be passed as part of the event post. If there's a
                conflict, the event-level ones will win.

        Returns EventHandlerKey to the handler which you can use to later remove
        the handler via ``remove_handler_by_key``.

        For example:

        .. code::

            my_handler = self.machine.events.add_handler('ev', self.test))

        Then later to remove all the handlers that a module added, you could:
        for handler in handler_list:
        ``events.remove_handler(my_handler)``
        """
        if event is None:
            raise AssertionError("Cannot pass event None.")
        if not self.machine.options['production']:
            if hasattr(self.machine, "switches") and event in self.machine.switches:
                self.raise_config_error('Switch name "{name}" name used as event handler for {handler}. '
                                        'Did you mean "{name}_active"?'.format(name=event, handler=handler), 1)
            if not callable(handler):
                raise AssertionError('Cannot add handler "{}" for event "{}". Did you '
                                     'accidentally add parenthesis to the end of the '
                                     'handler you passed?'.format(handler, event))

            sig = inspect.signature(handler)
            if 'kwargs' not in sig.parameters:
                raise AssertionError("Handler {} for event '{}' is missing **kwargs. Actual signature: {}".format(
                    handler, event, sig))

            if sig.parameters['kwargs'].kind != inspect.Parameter.VAR_KEYWORD:
                raise AssertionError("Handler {} for event '{}' param kwargs is missing '**'. "
                                     "Actual signature: {}".format(handler, event, sig))

        event, condition, additional_priority = self.get_event_and_condition_from_string(event)
        priority += additional_priority

        key = uuid.uuid4()

        # An event 'handler' in our case is a tuple with 4 elements:
        # the handler method, priority, dict of kwargs, & uuid key
        if hasattr(handler, "relative_priority") and not isinstance(handler, MagicMock):
            priority += handler.relative_priority

        self.registered_handlers[event].append(RegisteredHandler(handler, priority, kwargs, key, condition,
                                                                 blocking_facility))

        if self._debug:
            try:
                self.debug_log("Registered %s as a handler for '%s', priority: %s, "
                               "kwargs: %s",
                               (str(handler).split(' '))[2], event, priority, kwargs)
            except IndexError:
                self.debug_log("Registered %s as a handler for '%s', priority: %s, "
                               "kwargs: %s",
                               str(handler), event, priority, kwargs)

        # Sort the handlers for this event based on priority. We do it now
        # so the list is pre-sorted so we don't have to do that with each
        # event post.
        if len(self.registered_handlers[event]) > 1:
            self.registered_handlers[event].sort(key=lambda x: x.priority, reverse=True)

        if self._info:
            self._verify_handlers(event, self.registered_handlers[event])

        return EventHandlerKey(key, event)

    def _get_handler_signature(self, handler):
        """Perform black magic to calculate a signature for a handler."""
        cls = handler.callback.__self__

        # noinspection PyProtectedMember
        # pylint: disable-msg=protected-access
        if hasattr(self.machine, "device_manager") and cls == self.machine.device_manager and \
                handler.callback == self.machine.device_manager._control_event_handler:
            cls = (handler.kwargs["callback"].__self__, handler.kwargs["ms_delay"])

        handler_signature = (cls, handler.priority, handler.condition)
        return handler_signature

    def _verify_handlers(self, event, sorted_handlers):
        """Verify that no races can happen."""
        if not sorted_handlers or len(sorted_handlers) <= 1 or event.startswith("init_phase_"):
            return

        seen = set()
        collisions = []
        for handler in sorted_handlers:
            if not inspect.ismethod(handler.callback):
                continue

            handler_signature = self._get_handler_signature(handler)

            if handler_signature not in seen:
                seen.add(handler_signature)

            else:
                collisions.append(handler_signature)

        for collision in collisions:
            handlers = [x for x in sorted_handlers if inspect.ismethod(x.callback) and
                        self._get_handler_signature(x) == collision]
            self.info_log(
                "Unordered handler for class {} on event {} with priority {}. Handlers: {}. The order of those "
                "handlers is not defined and they will be executed in random order. This might lead to race "
                "conditions and potential bugs.".format(collision[0], event, collision[1], handlers)
            )

    def replace_handler(self, event: str, handler: Any, priority: int = 1,
                        **kwargs: dict) -> EventHandlerKey:
        """Check to see if a handler (optionally with kwargs) is registered for an event and replaces it if so.

        Args:
        ----
            event: The event you want to check to see if this handler is
                registered for.
            handler: The method of the handler you want to check.
            priority: Optional priority of the new handler that will be
                registered.
            **kwargs: The kwargs you want to check and the kwargs that will be
                registered with the new handler.

        If you don't pass kwargs, this method will just look for the handler and
        event combination. If you do pass kwargs, it will make sure they match
        before replacing the existing entry.

        If this method doesn't find a match, it will still add the new handler.
        """
        # Check to see if this handler is already registered for this event.
        # If we don't have kwargs, then we'll look for just the handler meth.
        # If we have kwargs, we'll look for that combination. If it finds it,
        # remove it.
        if event in self.registered_handlers:
            if kwargs:
                # slice the full list [:] to make a copy so we can delete from the
                # original while iterating
                for rh in self.registered_handlers[event][:]:
                    if rh[0] == handler and rh[2] == kwargs:
                        self.registered_handlers[event].remove(rh)
            else:
                for rh in self.registered_handlers[event][:]:
                    if rh[0] == handler:
                        self.registered_handlers[event].remove(rh)

        return self.add_handler(event, handler, priority, **kwargs)

    def remove_all_handlers_for_event(self, event: str) -> None:
        """Remove all handlers for event.

        Use carefully. This is currently used to remove handlers for all init events which only occur once.
        """
        if event in self.registered_handlers:
            del self.registered_handlers[event]

    def remove_handler(self, method: Any) -> None:
        """Remove an event handler from all events a method is registered to handle.

        Args:
        ----
            method : The method whose handlers you want to remove.
        """
        events_to_delete_if_empty = []
        for event, handler_list in self.registered_handlers.items():
            for handler_tup in handler_list[:]:  # copy via slice
                if handler_tup[0] == method:
                    handler_list.remove(handler_tup)
                    if self._debug:
                        self.debug_log("Removing method %s from event %s", (str(method).split(' '))[2], event)
                    events_to_delete_if_empty.append(event)

        for event in events_to_delete_if_empty:
            self._remove_event_if_empty(event)

    def remove_handler_by_event(self, event: str, handler: Any) -> None:
        """Remove the handler you pass from the event you pass.

        Args:
        ----
            event: The name of the event you want to remove the handler from.
            handler: The handler method you want to remove.

        Note that keyword arguments for the handler are not taken into
        consideration. In other words, this method only removes the registered
        handler / event combination, regardless of whether the keyword
        arguments match or not.
        """
        events_to_delete_if_empty = []
        if event in self.registered_handlers:
            for handler_tup in self.registered_handlers[event][:]:
                if handler_tup[0] == handler:
                    self.registered_handlers[event].remove(handler_tup)
                    if self._debug:
                        self.debug_log("Removing method %s from event %s", (str(handler).split(' '))[2], event)
                    events_to_delete_if_empty.append(event)

        for this_event in events_to_delete_if_empty:
            self._remove_event_if_empty(this_event)

    def remove_handler_by_key(self, key: EventHandlerKey) -> None:
        """Remove a registered event handler by key.

        Args:
        ----
            key: The key of the handler you want to remove
        """
        if key.event not in self.registered_handlers:
            return
        events_to_delete_if_empty = []
        for handler_tup in self.registered_handlers[key.event][:]:  # copy via slice
            if handler_tup.key == key.key:
                self.registered_handlers[key.event].remove(handler_tup)
                if self._debug:
                    self.debug_log("Removing method %s from event %s", (str(handler_tup[0]).split(' '))[2], key.event)
                events_to_delete_if_empty.append(key.event)
        for event in events_to_delete_if_empty:
            self._remove_event_if_empty(event)

    def remove_handlers_by_keys(self, key_list: List[EventHandlerKey]) -> None:
        """Remove multiple event handlers based on a passed list of keys.

        Args:
        ----
            key_list: A list of keys of the handlers you want to remove
        """
        for key in key_list:
            self.remove_handler_by_key(key)

    def _remove_event_if_empty(self, event: str) -> None:
        # Checks to see if the event doesn't have any more registered handlers,
        # removes it if so.

        if event not in self.registered_handlers:
            return

        if not self.registered_handlers[event]:  # if value is empty list
            del self.registered_handlers[event]
            if self._debug:
                self.debug_log("Removing event %s since there are no more"
                               " handlers registered for it", event)

    def wait_for_event(self, event_name: str) -> asyncio.Future:
        """Wait for event."""
        return self.wait_for_any_event([event_name])

    def wait_for_any_event(self, event_names: List[str]) -> asyncio.Future:
        """Wait for any event from event_names."""
        future = asyncio.Future()   # type: asyncio.Future
        keys = []   # type: List[EventHandlerKey]
        for event_name in event_names:
            keys.append(self.add_handler(event_name, partial(self._wait_handler,
                                                             _future=future,
                                                             _keys=keys,
                                                             event=event_name)))
        return future

    def _wait_handler(self, _future: asyncio.Future, _keys: List[EventHandlerKey], **kwargs):
        for key in _keys:
            self.remove_handler_by_key(key)

        if _future.cancelled():
            return
        _future.set_result(kwargs)

    def does_event_exist(self, event_name: str) -> bool:
        """Check to see if any handlers are registered for the event name that is passed.

        Args:
        ----
            event_name : The string name of the event you want to check.

        Returns True or False.
        """
        return event_name in self.registered_handlers

    @staticmethod
    def _set_result(_future, **kwargs):
        if not _future.done():
            _future.set_result(kwargs)

    def post_async(self, event: str, **kwargs: dict) -> asyncio.Future:
        """Post event and wait until all handlers are done."""
        future = asyncio.Future()   # type: asyncio.Future
        self.post(event, partial(self._set_result, _future=future), **kwargs)
        return future

    def post_relay_async(self, event: str, **kwargs: dict) -> asyncio.Future:
        """Post relay event, wait until all handlers are done and return result."""
        future = asyncio.Future()   # type: asyncio.Future
        self.post_relay(event, partial(self._set_result, _future=future), **kwargs)
        return future

    def post_queue_async(self, event: str, **kwargs: dict) -> asyncio.Future:
        """Post queue event, wait until all handlers are done and locks are released."""
        future = asyncio.Future()   # type: asyncio.Future
        self.post_queue(event, partial(self._set_result, _future=future), **kwargs)
        return future

    def post(self, event: str, callback=None, **kwargs) -> None:
        """Post an event which causes all the registered handlers to be called.

        Events are processed serially (e.g. one at a time), so if the event
        core is in the process of handling another event, this event is
        added to a queue and processed after the current event is done.

        You can control the order the handlers will be called by optionally
        specifying a priority when the handlers were registered. (Higher
        priority values will be processed first.)

        Args:
        ----
            event: A string name of the event you're posting. Note that you can
                post whatever event you want. You don't have to set up anything
                ahead of time, and if no handlers are registered for the event
                you post, so be it.
            callback: An optional method which will be called when the final
                handler is done processing this event. Default is None.
            **kwargs: One or more options keyword/value pairs that will be
                passed to each handler. (The event manager will enforce that
                handlers have ``**kwargs`` in their signatures when they're
                registered to prevent run-time crashes from unexpected kwargs
                that were included in ``post()`` calls.
        """
        self._post(event, None, callback, **kwargs)

    def post_boolean(self, event: str, callback=None, **kwargs) -> None:
        """Post an boolean event which causes all the registered handlers to be called one-by-one.

        Boolean events differ from regular events in that if any handler
        returns False, the remaining handlers will not be called.

        Events are processed serially (e.g. one at a time), so if the event
        core is in the process of handling another event, this event is
        added to a queue and processed after the current event is done.

        You can control the order the handlers will be called by optionally
        specifying a priority when the handlers were registered. (Higher
        priority values will be processed first.)

        Args:
        ----
            event: A string name of the event you're posting. Note that you can
                post whatever event you want. You don't have to set up anything
                ahead of time, and if no handlers are registered for the event
                you post, so be it.
            callback: An optional method which will be called when the final
                handler is done processing this event. Default is None. If
                any handler returns False and cancels this boolean event, the
                callback will still be called, but a new kwarg ev_result=False
                will be passed to it.
            **kwargs: One or more options keyword/value pairs that will be
                passed to each handler.
        """
        self._post(event, 'boolean', callback, **kwargs)

    def post_queue(self, event, callback, **kwargs):
        """Post a queue event which causes all the registered handlers to be called.

        Queue events differ from standard events in that individual handlers
        are given the option to register a "wait", and the callback will not be
        called until any handler(s) that registered a wait will have to release
        that wait. Once all the handlers release their waits, the callback is
        called.

        Events are processed serially (e.g. one at a time), so if the event
        core is in the process of handling another event, this event is
        added to a queue and processed after the current event is done.

        You can control the order the handlers will be called by optionally
        specifying a priority when the handlers were registered. (Higher
        numeric values will be processed first.)

        Args:
        ----
        ----
            event: A string name of the event you're posting. Note that you can
                post whatever event you want. You don't have to set up anything
                ahead of time, and if no handlers are registered for the event
                you post, so be it.
            callback: The method which will be called when the final
                handler is done processing this event and any handlers that
                registered waits have cleared their waits.
            **kwargs: One or more options keyword/value pairs that will be
                passed to each handler. (Just make sure your handlers are
                expecting them. You can add ``**kwargs`` to your handler
                methods if certain ones don't need them.)

        Example:
        -------
        Post the queue event called *pizza_time*, and then call
        ``self.pizza_done`` when done:

        .. code::

             self.machine.events.post_queue('pizza_time', self.pizza_done)

        """
        self._post(event, 'queue', callback, **kwargs)

    def post_relay(self, event: str, callback=None, **kwargs) -> None:
        """Post a relay event which causes all the registered handlers to be called.

        A dictionary can be passed from handler-to-handler and modified
        as needed.

        Args:
        ----
            event: A string name of the event you're posting. Note that you can
                post whatever event you want. You don't have to set up anything
                ahead of time, and if no handlers are registered for the event
                you post, so be it.
            callback: The method which will be called when the final handler is
                done processing this event. Default is None.
            **kwargs: One or more options keyword/value pairs that will be
                passed to each handler. (Just make sure your handlers are
                expecting them. You can add ``**kwargs`` to your handler
                methods if certain ones don't need them.)

        Events are processed serially (e.g. one at a time), so if the event
        core is in the process of handling another event, this event is
        added to a queue and processed after the current event is done.

        You can control the order the handlers will be called by optionally
        specifying a priority when the handlers were registered. (Higher
        priority values will be processed first.)

        Relay events differ from standard events in that the resulting kwargs
        from one handler are passed to the next handler. (In other words,
        standard events mean that all the handlers get the same initial kwargs,
        whereas relay events "relay" the resulting kwargs from one handler to
        the next.)
        """
        self._post(event, 'relay', callback, **kwargs)

    def _post(self, event: str, ev_type: Optional[str], callback, **kwargs: dict) -> None:
        if self._debug:
            self.debug_log("Event: ===='%s'==== Type: %s, Callback: %s, "
                           "Args: %s", event, ev_type, callback, kwargs)
        elif self._info and not kwargs.get("_silent", False):
            self.info_log("Event: ======'%s'====== Args=%s", event, kwargs)

        # fast path for events without handler
        if not callback and not self.monitor_events and event not in self.registered_handlers:
            return

        if not self.event_queue and hasattr(self.machine.clock, "loop"):
            self.machine.clock.loop.call_soon(self.process_event_queue)

        posted_event = PostedEvent(event, ev_type, callback, kwargs)

        if self.monitor_events and not kwargs.get("_silent", False):
            self.machine.bcp.interface.monitor_posted_event(posted_event)

        self.event_queue.append(posted_event)
        if self._debug:
            self.debug_log("+============= EVENTS QUEUE =============")
            for this_event in list(self.event_queue):    # type: ignore
                self.debug_log("| %s, %s, %s, %s", this_event[0], this_event[1],
                               this_event[2], this_event[3])
            self.debug_log("+========================================")

    async def _run_handlers_sequential(self, event: str, callback, kwargs: dict) -> None:
        """Run all handlers for an event."""
        if self._debug:
            self.debug_log("^^^^ Processing queue event '%s'. Callback: %s,"
                           " Args: %s", event, callback, kwargs)

        # all handlers may have been removed in the meantime
        if event not in self.registered_handlers:
            return

        # Now let's call the handlers one-by-one, including any kwargs
        for handler in self.registered_handlers[event][:]:
            # use slice above so we don't process new handlers that came
            # in while we were processing previous handlers

            # merge the post's kwargs with the registered handler's kwargs
            # in case of conflict, handlers kwargs will win
            merged_kwargs = dict(list(kwargs.items()) + list(handler.kwargs.items()))

            # if condition exists and is not true skip
            if handler.condition is not None and not handler.condition.evaluate(merged_kwargs):
                continue

            # log if debug is enabled and this event is not the timer tick
            if self._debug:
                try:
                    self.debug_log("%s (priority: %s) responding to event '%s'"
                                   " with args %s",
                                   (str(handler.callback).split(' ')), handler.priority,
                                   event, merged_kwargs)
                except IndexError:
                    pass

            # call the handler and save the results

            try:
                queue = merged_kwargs.pop('queue')
            except KeyError:
                queue = QueuedEvent(self.debug_log)

            handler.callback(queue=queue, **merged_kwargs)

            if queue.waiter:
                queue.event = asyncio.Event()
                await queue.event.wait()

        if self._debug:
            self.debug_log("vvvv Finished queue event '%s'. Callback: %s. "
                           "Args: %s", event, callback, kwargs)

        if callback:
            callback(**kwargs)

    def _run_handlers(self, event: str, ev_type: Optional[str], kwargs: dict) -> Any:
        """Run all handlers for an event."""
        result = None
        for handler in self.registered_handlers[event][:]:
            # use slice above so we don't process new handlers that came
            # in while we were processing previous handlers

            if '_min_priority' in kwargs and handler.blocking_facility and \
                (kwargs['_min_priority']['all'] > handler.priority or (
                    handler.blocking_facility in kwargs['_min_priority'] and
                    kwargs['_min_priority'][handler.blocking_facility] > handler.priority)):
                continue

            if handler.kwargs and kwargs:
                # merge the post's kwargs with the registered handler's kwargs
                # in case of conflict, handler kwargs will win
                merged_kwargs = dict(list(kwargs.items()) + list(handler.kwargs.items()))
            elif handler.kwargs:
                merged_kwargs = handler.kwargs
            else:
                merged_kwargs = kwargs

            # if condition exists and is not true skip
            if handler.condition is not None and not handler.condition.evaluate(merged_kwargs):
                continue

            if self._debug:
                try:
                    self.debug_log("%s (priority: %s) responding to event '%s'"
                                   " with args %s",
                                   (str(handler.callback).split(' ')), handler.priority,
                                   event, merged_kwargs)
                except IndexError:
                    pass

            # call the handler and save the results
            try:
                result = handler.callback(**merged_kwargs)
            except Exception as e:
                raise Exception("Exception while processing {} for event {}. {}".format(handler, event, e)) from e

            # If whatever handler we called returns False, we stop
            # processing the remaining handlers for boolean or queue events
            if ev_type == 'boolean' and result is False:
                # add a False result so our callback knows something failed
                kwargs['ev_result'] = False

                if self._debug:
                    self.debug_log("Aborting future event processing")
                break

            if ev_type == 'relay' and isinstance(result, dict):
                kwargs.update(result)
            elif isinstance(result, dict) and '_min_priority' in result:
                kwargs['_min_priority'] = result['_min_priority']

        return result

    def _process_queue_event(self, event: str, callback, **kwargs: dict):
        """Handle queue events."""
        if event not in self.registered_handlers:
            # fast path if there are not handlers
            self.callback_queue.append((callback, kwargs))
        else:
            task = self.machine.clock.loop.create_task(self._run_handlers_sequential(event, callback, kwargs))
            task.add_done_callback(self._queue_task_done)
            self._queue_tasks.append(task)

    def _queue_task_done(self, future):
        """Remove queue task from list and evaluate result."""
        future.result()
        self._queue_tasks.remove(future)

    def _process_event(self, event: str, ev_type: Optional[str], callback=None, **kwargs: dict) -> None:
        # Internal method which actually handles the events. Don't call this.

        result = None
        if self._debug:
            self.debug_log("^^^^ Processing event '%s'. Type: %s, Callback: %s,"
                           " Args: %s", event, ev_type, callback, kwargs)

        # Now let's call the handlers one-by-one, including any kwargs
        if event in self.registered_handlers:
            result = self._run_handlers(event, ev_type, kwargs)

        if self._debug:
            self.debug_log("vvvv Finished event '%s'. Type: %s. Callback: %s. "
                           "Args: %s", event, ev_type, callback, kwargs)

        if callback:
            # For event types other than queue, we'll handle the callback here.
            # Queue events with active waits will do the callback when the
            # waits clear

            if result:
                # if our last handler returned something, add it to kwargs
                kwargs['ev_result'] = result

            self.callback_queue.append((callback, kwargs))

    def process_event_queue(self) -> None:
        """Check if there are any other events that need to be processed, and then process them."""
        inner_queue = deque()   # type: Deque[Deque[PostedEvent]]
        while self.event_queue or self.callback_queue:
            # first process all events. if they post more events we will
            # process them in the same loop.
            if self.event_queue:
                next_queue = self.event_queue
                self.event_queue = deque()
                while next_queue:
                    # remember the previous queue since events might be posted in this handler

                    event = next_queue.popleft()
                    if not next_queue and inner_queue:
                        next_queue = inner_queue.popleft()

                    if event.type == "queue":
                        self._process_queue_event(event=event[0],
                                                  callback=event[2],
                                                  **event[3])
                    else:
                        self._process_event(event=event[0],
                                            ev_type=event[1],
                                            callback=event[2],
                                            **event[3])

                    # make sure the handler created during this handler are called first
                    if self.event_queue:
                        inner_queue.appendleft(next_queue)
                        next_queue = self.event_queue
                        self.event_queue = deque()

            # when all events are processed run the _last_ callback. afterwards
            # continue with the loop and run all events. this makes sure all
            # events are completed before running the callback
            if self.callback_queue:
                callback, kwargs = self.callback_queue.pop()
                callback(**kwargs)


class QueuedEvent:

    """Base class for an event queue which is created each time a queue event is called."""

    def __init__(self, debug_log: Callable[[str], None]) -> None:
        """Initialize QueueEvent."""
        self.debug_log = debug_log
        self.waiter = False
        self.event = None   # type: Optional[asyncio.Event]

    def __repr__(self):
        """Return str representation."""
        return '<QueuedEvent>'

    def wait(self) -> None:
        """Register a wait for this QueueEvent."""
        if self.waiter:
            raise AssertionError("Double lock")
        self.waiter = True
        self.debug_log("QueuedEvent: Registering a wait.")

    def clear(self) -> None:
        """Clear a wait."""
        if not self.waiter:
            raise AssertionError("Not locked")

        self.waiter = False

        # in case this is async we release the lock
        if self.event:
            self.event.set()

    def is_empty(self) -> bool:
        """Return true if unlocked."""
        return not self.waiter


def event_handler(relative_priority):
    """Decorate an event handler."""
    def decorator(func):
        """Decorate a function with relative priority."""
        func.relative_priority = relative_priority
        return func

    return decorator



"""An achievement group which manages and groups achievements."""
from typing import Optional, List

from random import choice

from mpf.core.events import event_handler, EventHandlerKey
from mpf.core.machine import MachineController
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player
from mpf.core.device_monitor import DeviceMonitor

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.achievement import Achievement     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.assets.show import RunningShow     # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor(_enabled="enabled", _selected_member="selected_member")
class AchievementGroup(ModeDevice):

    """An achievement group in a pinball machine.

    It is tracked per player and can automatically restore state on the next
    ball.
    """

    __slots__ = ["_show", "_enabled", "_selected_member", "_rotation_in_progress", "_handlers", "_loaded"]

    config_section = 'achievement_groups'
    collection = 'achievement_groups'
    class_label = 'achievement_group'

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialize achievement."""
        super().__init__(machine, name)

        self._show = None       # type: Optional[RunningShow]

        self._enabled = False
        self._loaded = False
        self._selected_member = None    # type: Optional[Achievement]
        self._rotation_in_progress = False
        self._handlers = []             # type: List[EventHandlerKey]

    @property
    def enabled(self):
        """Return enabled state."""
        return self._enabled

    def enable(self):
        """Enable achievement group."""
        if not self._loaded:
            return

        if self._enabled:
            self.debug_log("Group is already enabled. Aborting")
            return

        if self._is_member_started() and self.config['disable_while_achievement_started']:
            self.debug_log("Not enabling because a member is started and disable_while_achievement_started is true.")
            return

        super().enable()
        self.debug_log("Call to enable this group")

        self._enabled = True
        self.debug_log("Enabling group")

        self._stop_show()
        if self._selected_member and self._selected_member.selected:
            self._selected_member = None

        show = self.config['show_when_enabled']

        if show:
            self._show = show.play(
                priority=self.mode.priority,
                loops=-1,
                show_tokens=self.config['show_tokens'])

        for e in self.config['events_when_enabled']:
            self.machine.events.post(e)

        self._process_current_member_state()

    @event_handler(0)
    def event_disable(self, **kwargs):
        """Event handler for disable event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable achievement group."""
        if not self._enabled:
            self.debug_log("Group is already disabled. Aborting")
            return
        self.debug_log("Disabling group")
        self._stop_show()
        self._enabled = False

    def _stop_show(self):
        if self._show:
            self.debug_log("Stopping show")
            self._show.stop()
            self._show = None

    def _get_available_achievements_for_selection(self):
        """Return achievements which can be selected and started."""
        return [x for x in self.config['achievements'] if x.can_be_selected_for_start]

    def _get_current(self):
        self.debug_log("Getting current selected achievement")
        if not self._selected_member:
            self.select_random_achievement()

        return self._selected_member

    @event_handler(5)
    def event_start_selected(self, **kwargs):
        """Event handler for start_selected event."""
        del kwargs
        self.start_selected()

    def start_selected(self):
        """Start the currently selected achievement."""
        self.debug_log("Call to start selected achievement")
        if not self._enabled:
            self.debug_log("Group not enabled. Aborting")
            return

        try:
            self._get_current().start()
        except AttributeError:
            # don't have a current one
            pass

    @event_handler(6)
    def event_rotate_right(self, reverse=False, **kwargs):
        """Event handler for rotate_right event."""
        del kwargs
        self.rotate_right(reverse)

    def rotate_right(self, reverse=False):
        """Rotate to the right."""
        self.debug_log("Call to rotate")
        if not self._is_ok_to_change_selection():
            return

        if not self._selected_member and not self.config['auto_select']:
            self.debug_log("Nothing selected and auto_select false. Abort.")
            return

        self._rotation_in_progress = True

        # if there's already one selected, set it back to enabled
        if self._selected_member and self._selected_member.selected:
            self._selected_member.unselect()

        achievements = self._get_available_achievements_for_selection()
        if not achievements:
            # there is nothing to rotate
            self.debug_log("Nothing to rotate. Abort.")
            return

        try:
            current_index = achievements.index(self._get_current())
        except ValueError:
            self._selected_member = self._get_current()
        else:
            if reverse:
                self._selected_member = achievements[(current_index - 1) % len(achievements)]
            else:
                self._selected_member = achievements[(current_index + 1) % len(achievements)]

        self._selected_member.select()

        self._rotation_in_progress = False

    @event_handler(7)
    def event_rotate_left(self, **kwargs):
        """Event handler for rotate_left event."""
        del kwargs
        self.rotate_left()

    def rotate_left(self):
        """Rotate to the left."""
        self.rotate_right(reverse=True)

    def _no_more_enabled(self):
        """Post event when no more enabled achievements are available."""
        self.debug_log("No more achievements are enabled")
        for e in self.config['events_when_no_more_enabled']:
            self.machine.events.post(e)

    def _all_complete(self):
        self.debug_log("All achievements are complete")
        self.disable()  # disable before event post so event can re-enable

        for e in self.config['events_when_all_completed']:
            self.machine.events.post(e)

    @event_handler(9)
    def event_select_random_achievement(self, **kwargs):
        """Event handler for select_random_achievement event."""
        del kwargs
        self.select_random_achievement()

    def select_random_achievement(self):
        """Select a random achievement."""
        self.debug_log("Selecting random achievement")

        if not self._is_ok_to_change_selection():
            self.debug_log("Not ok to change selection. Aborting")
            return

        if self._selected_member and self._selected_member.selected:
            self._selected_member.unselect()

        try:
            ach = choice(self._get_available_achievements_for_selection())
            # todo change this to use our Randomizer class
            self._selected_member = ach
            self.debug_log("Picked new random achievement: %s", ach)
            ach.select()
        except IndexError:
            self._no_more_enabled()

    def _is_ok_to_change_selection(self):
        self.debug_log("Checking if it's ok to change selection...")
        if self._enabled:
            self.debug_log("Ok because enabled")
            return True
        if self.config['allow_selection_change_while_disabled']:
            self.debug_log("Ok because allow_selection_change_while_disabled is set (but disabled)")
            return True

        self.debug_log("not ok")
        return False

    def member_state_changed(self, **kwargs):
        """Notify the group that one of its members has changed state."""
        del kwargs
        if not self._loaded:
            return
        self.debug_log("Member state has changed")
        if self._is_member_started():
            self.debug_log("A member is started")
            if self.config['disable_while_achievement_started'] and self.enabled:
                self.debug_log("disable_while_achievement_started is true")
                self.disable()
            else:
                self._process_current_member_state()

        elif self.config['enable_while_no_achievement_started'] and not self.enabled:
            self.debug_log("enable_while_no_achievement_started is true")
            self.enable()
        else:
            self._process_current_member_state()

    def _process_current_member_state(self):
        self.debug_log("Processing current member state")

        if self._rotation_in_progress:
            self.debug_log("Rotation in progress. Aborting")
            return

        if not self._enabled:
            self.debug_log("Not enabled. Aborting")
            return

        self._check_for_all_complete()
        if self._check_for_no_more_enabled():
            self.debug_log("No achievement enabled. Aborting.")
            return
        self._update_selected()

        if not self._selected_member and self.config['auto_select']:
            self.debug_log("No selected member, but auto_select is true")
            self.select_random_achievement()

    def _update_selected(self):
        self.debug_log("Updating selected achievement")
        for ach in self.config['achievements']:
            if ach.selected:
                self._selected_member = ach
                self.debug_log("Already have a selected member is %s", ach)
                return True

            self.debug_log("Do not have a current selected member")

        if self.config['auto_select']:
            self.debug_log("Auto select is true. Getting random achievement")
            self.select_random_achievement()

        return False

    def _check_for_all_complete(self):
        self.debug_log("Checking for all complete")
        if not self._enabled:
            self.debug_log("Group is not enabled. Aborting...")
            return False

        if not [x for x in self.config['achievements'] if x.state != "completed"]:
            self._all_complete()
            return True
        self.debug_log("All are not complete")
        return False

    def _check_for_no_more_enabled(self):
        self.debug_log("Checking for no more enabled")

        if not self._get_available_achievements_for_selection():
            self._no_more_enabled()
            return True
        return False

    def _is_member_started(self):
        self.debug_log("Checking if member is started")
        for ach in self.config['achievements']:
            if ach.state == 'started':
                self.debug_log("Found %s is started", ach)
                return True
        self.debug_log("No member is started")
        return False

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Load device on mode start and restore state.

        Args:
        ----
            mode: mode which was contains the device
            player: player which is currently active
        """
        super().device_loaded_in_mode(mode, player)
        self._loaded = True

        for ach in self.config['achievements']:
            self._handlers.append(self.machine.events.add_handler("achievement_{}_changed_state".format(ach.name),
                                                                  self.member_state_changed))

        if self.config['enable_while_no_achievement_started'] and self._is_member_started() and not self._enabled:
            self.enable()

    def device_removed_from_mode(self, mode: Mode):
        """Mode ended.

        Args:
        ----
            mode: mode which stopped
        """
        super().device_removed_from_mode(mode)

        self.machine.events.remove_handlers_by_keys(self._handlers)
        self._handlers = []

        self.disable()
        self._loaded = False
        self._selected_member = None

        if self._show:
            self._show.stop()

    def _enable_related_device_debugging(self):
        for ach in self.config['achievements']:
            ach.enable_debugging()



"""OPP input card."""
import logging

from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface

from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf


class OPPInputCard:

    """OPP input card."""

    __slots__ = ["log", "chain_serial", "addr", "is_matrix", "old_state", "mask", "card_num"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, chain_serial, addr, mask, inp_dict, inp_addr_dict):
        """Initialise OPP input card."""
        self.log = logging.getLogger('OPPInputCard {} on {}'.format(addr, chain_serial))
        self.chain_serial = chain_serial
        self.addr = addr
        self.is_matrix = False
        self.old_state = 0
        self.mask = mask
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))

        self.log.debug("Creating OPP Input at hardware address: 0x%02x", addr)

        inp_addr_dict[chain_serial + '-' + str(addr)] = self
        for index in range(0, 32):
            if ((1 << index) & mask) != 0:
                inp_dict[self.chain_serial + "-" + self.card_num + '-' + str(index)] =\
                    OPPSwitch(self, self.chain_serial + "-" + self.card_num + '-' + str(index))


class OPPMatrixCard:

    """OPP matrix input card."""

    __slots__ = ["log", "chain_serial", "addr", "mask", "is_matrix", "old_state", "card_num"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, chain_serial, addr, inp_dict, inp_addr_dict):
        """Initialise OPP matrix input card."""
        self.log = logging.getLogger('OPPMatrixCard')
        self.chain_serial = chain_serial
        self.addr = addr
        self.mask = 0xFFFFFFFFFFFFFFFF << 32  # create fake mask
        self.is_matrix = True
        self.old_state = [0, 0]
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))

        self.log.debug("Creating OPP Matrix Input at hardware address: 0x%02x", addr)

        inp_addr_dict[chain_serial + '-' + str(addr)] = self

        # Matrix inputs are inputs 32 - 95 (OPP only supports 8x8 input switch matrices)
        for index in range(32, 96):
            inp_dict[self.chain_serial + "-" + self.card_num + '-' + str(index)] =\
                OPPSwitch(self, self.chain_serial + "-" + self.card_num + '-' + str(index))


class OPPSwitch(SwitchPlatformInterface):

    """An OPP input on an OPP input card."""

    __slots__ = ["card"]

    def __init__(self, card, number):
        """Initialise input."""
        super().__init__({}, number)
        self.card = card

    def get_board_name(self):
        """Return OPP chain and addr."""
        return "OPP {} Board {}".format(str(self.card.chain_serial), "0x%02x" % self.card.addr)



"""A light system for platforms which batches all updates."""
import abc
import asyncio

from typing import Tuple, Set, List
from sortedcontainers import SortedSet, SortedList
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.core.utility_functions import Util


class PlatformBatchLight(LightPlatformInterface, abc.ABC):

    """Light which can be batched."""

    __slots__ = ["light_system", "_current_fade", "_last_brightness"]

    def __init__(self, number, light_system: "PlatformBatchLightSystem"):
        """Initialise light."""
        super().__init__(number)
        self.light_system = light_system
        self._current_fade = (0, -1, 0, -1)
        self._last_brightness = None

    @abc.abstractmethod
    def get_max_fade_ms(self):
        """Return max fade ms."""

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Mark dirty and remember fade."""
        self.light_system.mark_dirty(self)
        self._current_fade = (start_brightness, start_time, target_brightness, target_time)
        self._last_brightness = None

    def get_fade_and_brightness(self, current_time):
        """Return fade + brightness and mark as clean if this is it."""
        if self._last_brightness:
            return self._last_brightness, 0, True
        max_fade_ms = self.get_max_fade_ms()
        start_brightness, start_time, target_brightness, target_time = self._current_fade
        fade_ms = int(round((target_time - current_time) * 1000.0))
        if fade_ms > max_fade_ms > 0:
            fade_ms = max_fade_ms
            ratio = ((current_time + (fade_ms / 1000.0) - start_time) /
                     (target_time - start_time))
            brightness = start_brightness + (target_brightness - start_brightness) * ratio
            done = False
        else:
            if fade_ms < 0:
                fade_ms = 0
            brightness = target_brightness
            self._last_brightness = brightness
            done = True

        return brightness, fade_ms, done


class PlatformBatchLightSystem:

    """Batch light system for platforms."""

    __slots__ = ["dirty_lights", "dirty_schedule", "clock", "update_task", "update_callback",
                 "update_hz", "max_batch_size", "scheduler_task", "schedule_changed", "dirty_lights_changed",
                 "last_state"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, clock, update_callback, update_hz, max_batch_size):
        """Initialise light system."""
        self.dirty_lights = SortedSet()    # type: Set[PlatformBatchLight]
        self.dirty_lights_changed = asyncio.Event()
        self.dirty_schedule = SortedList()
        self.schedule_changed = asyncio.Event()
        self.update_task = None
        self.scheduler_task = None
        self.clock = clock
        self.update_callback = update_callback
        self.update_hz = update_hz
        self.max_batch_size = max_batch_size
        self.last_state = {}

    def start(self):
        """Start light system."""
        self.update_task = self.clock.loop.create_task(self._send_updates())
        self.update_task.add_done_callback(Util.raise_exceptions)
        self.scheduler_task = self.clock.loop.create_task(self._schedule_updates())
        self.scheduler_task.add_done_callback(Util.raise_exceptions)

    def stop(self):
        """Stop light system."""
        if self.scheduler_task:
            self.scheduler_task.cancel()
            self.scheduler_task = None
        if self.update_task:
            self.update_task.cancel()
            self.update_task = None

    async def _schedule_updates(self):
        while True:
            run_time = self.clock.get_time()
            self.schedule_changed.clear()
            while self.dirty_schedule and self.dirty_schedule[0][0] <= run_time:
                self.dirty_lights.add(self.dirty_schedule[0][1])
                del self.dirty_schedule[0]
            self.dirty_lights_changed.set()

            if self.dirty_schedule:
                await asyncio.wait([self.schedule_changed.wait()],
                                   timeout=self.dirty_schedule[0][0] - run_time, return_when=asyncio.FIRST_COMPLETED)
            else:
                await self.schedule_changed.wait()

    async def _send_updates(self):
        poll_sleep_time = 1 / self.update_hz
        max_fade_tolerance = int(poll_sleep_time * 1000)
        while True:
            await self.dirty_lights_changed.wait()
            self.dirty_lights_changed.clear()
            sequential_lights = []
            for light in list(self.dirty_lights):
                if not sequential_lights:
                    # first light
                    sequential_lights = [light]
                elif light.is_successor_of(sequential_lights[-1]):
                    # lights are sequential
                    sequential_lights.append(light)
                else:
                    # sequence ended
                    await self._send_update_batch(sequential_lights, max_fade_tolerance)
                    # this light is a new sequence
                    sequential_lights = [light]

            if sequential_lights:
                await self._send_update_batch(sequential_lights, max_fade_tolerance)

            self.dirty_lights.clear()

            await asyncio.sleep(poll_sleep_time)

    async def _send_update_batch(self, sequential_lights: List[PlatformBatchLight], max_fade_tolerance):
        sequential_brightness_list = []     # type: List[Tuple[LightPlatformInterface, float, int]]
        common_fade_ms = None
        current_time = self.clock.get_time()
        for light in sequential_lights:
            brightness, fade_ms, done = light.get_fade_and_brightness(current_time)
            schedule_time = current_time + (fade_ms / 1000)
            if not done:
                if not self.dirty_schedule or self.dirty_schedule[0][0] > schedule_time:
                    self.schedule_changed.set()
                self.dirty_schedule.add((schedule_time, light))
            else:
                # check if we realized this brightness earlier
                last_state = self.last_state.get(light, None)
                if last_state and last_state[0] == brightness and last_state[1] < schedule_time:
                    # we already set the light to that color earlier. skip it
                    if not sequential_brightness_list:
                        # we only skip this light if we are in the beginning of the list for now
                        # the reason for that is that we do not want to break fade chains when one color channel
                        # of an RGB light did not change
                        # this could become an option in the future
                        continue

            self.last_state[light] = (brightness, schedule_time)

            if common_fade_ms is None:
                common_fade_ms = fade_ms

            if -max_fade_tolerance < common_fade_ms - fade_ms < max_fade_tolerance and \
                    len(sequential_brightness_list) < self.max_batch_size:
                sequential_brightness_list.append((light, brightness, common_fade_ms))
            else:
                await self.update_callback(sequential_brightness_list)
                # start new list
                current_time = self.clock.get_time()
                common_fade_ms = fade_ms
                sequential_brightness_list = [(light, brightness, common_fade_ms)]

        if sequential_brightness_list:
            await self.update_callback(sequential_brightness_list)

    def mark_dirty(self, light: "PlatformBatchLight"):
        """Mark as dirty."""
        self.dirty_lights.add(light)
        self.dirty_lights_changed.set()
        self.dirty_schedule = SortedList([x for x in self.dirty_schedule if x[1] != light])



"""Contains the MultiBall device class."""
from mpf.core.enable_disable_mixin import EnableDisableMixin

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.mode_device import ModeDevice
from mpf.core.placeholder_manager import NativeTypeTemplate
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("shoot_again", "balls_added_live", "balls_live_target")
class Multiball(EnableDisableMixin, SystemWideDevice, ModeDevice):

    """Multiball device for MPF."""

    config_section = 'multiballs'
    collection = 'multiballs'
    class_label = 'multiball'

    __slots__ = ["ball_locks", "source_playfield", "delay", "balls_added_live", "balls_live_target", "shoot_again"]

    def __init__(self, machine, name):
        """Initialise multiball."""
        self.ball_locks = None
        self.source_playfield = None
        super().__init__(machine, name)

        self.delay = DelayManager(machine)
        self.balls_added_live = 0
        self.balls_live_target = 0
        self.shoot_again = False

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    def device_removed_from_mode(self, mode):
        """Disable and stop mb when mode stops."""
        super().device_removed_from_mode(mode)

        # also stop mb if shoot again is specified (aka the MB is currently running)
        if self.shoot_again:
            self.stop()

    async def _initialize(self):
        await super()._initialize()
        self.ball_locks = self.config['ball_locks']
        self.source_playfield = self.config['source_playfield']
        if isinstance(self.config['ball_count'], NativeTypeTemplate):
            ball_count = self.config['ball_count'].evaluate([])
        else:
            ball_count = None

        if self.config['ball_count_type'] == "total" and ball_count is not None and \
                ball_count <= 1:
            self.raise_config_error("ball_count should be at least 2 for a multiball to have an effect when "
                                    "ball_count_type is set to total.", 1)
        elif self.config['ball_count_type'] == "add" and ball_count is not None and \
                ball_count <= 0:
            self.raise_config_error("ball_count should be at least 1 for a multiball to have an effect when "
                                    "ball_count_type is set to add.", 2)

    @classmethod
    def prepare_config(cls, config, is_mode_config):
        """Add default enable_events and disable_events outside mode."""
        if not is_mode_config:
            if 'enable_events' not in config:
                config['enable_events'] = 'ball_started'
            if 'disable_events' not in config:
                config['disable_events'] = 'ball_will_end'
        return super().prepare_config(config, is_mode_config)

    def _handle_balls_in_play_and_balls_live(self):
        ball_count = self.config['ball_count'].evaluate([])
        balls_to_replace = self.machine.game.balls_in_play if self.config['replace_balls_in_play'] else 0
        self.debug_log("Going to add an additional {} balls for replace_balls_in_play".format(balls_to_replace))

        if self.config['ball_count_type'] == "total":
            # policy: total balls
            if ball_count > self.machine.game.balls_in_play:
                self.balls_added_live = ball_count - self.machine.game.balls_in_play
                self.machine.game.balls_in_play = ball_count
            self.balls_live_target = ball_count
        else:
            # policy: add balls
            self.balls_added_live = ball_count
            self.machine.game.balls_in_play += self.balls_added_live
            self.balls_live_target = self.machine.game.balls_in_play

        self.balls_added_live += balls_to_replace

    @event_handler(10)
    def event_start(self, **kwargs):
        """Event handler for start event."""
        del kwargs
        self.start()

    def start(self):
        """Start multiball."""
        if not self.enabled:
            return

        if self.balls_live_target > 0:
            self.debug_log("Cannot start MB because %s are still in play",
                           self.balls_live_target)
            return

        self.shoot_again = True

        self._handle_balls_in_play_and_balls_live()
        self.debug_log("Starting multiball with %s balls (added %s)", self.balls_live_target, self.balls_added_live)

        balls_added = 0

        # eject balls from locks
        for device in self.ball_locks:
            balls_to_release = max(min(device.available_balls, self.balls_added_live - balls_added), 0)
            device.eject(balls_to_release)
            balls_added += balls_to_release

        # request remaining balls
        if self.balls_added_live - balls_added > 0:
            self.source_playfield.add_ball(balls=self.balls_added_live - balls_added)

        shoot_again_ms = self.config['shoot_again'].evaluate([])
        if not shoot_again_ms:
            # No shoot again. Just stop multiball right away
            self.stop()
        else:
            # Enable shoot again
            self.machine.events.add_handler('ball_drain',
                                            self._ball_drain_shoot_again,
                                            priority=1000)
            # Register stop handler
            if shoot_again_ms > 0:
                self.delay.add(name='disable_shoot_again',
                               ms=shoot_again_ms,
                               callback=self.stop)

        self.machine.events.post("multiball_" + self.name + "_started",
                                 balls=self.balls_live_target)
        '''event: multiball_(name)_started
        desc: The multiball called (name) has just started.
        args:
            balls: The number of balls in this multiball
        '''

    def _ball_drain_shoot_again(self, balls, **kwargs):
        del kwargs

        balls_to_safe = self.balls_live_target - self.machine.game.balls_in_play + balls

        if balls_to_safe <= 0:
            return {'balls': balls}

        if balls_to_safe > balls:
            balls_to_safe = balls

        self.machine.events.post("multiball_" + self.name + "_shoot_again", balls=balls_to_safe)
        '''event: multiball_(name)_shoot_again
        desc: A ball has drained during the multiball called (name) while the
        ball save timer for that multiball was running, so a ball (or balls)
        will be saved and re-added into play.

        args:
            balls: The number of balls that are being saved.
        '''

        self.debug_log("Ball drained during MB. Requesting a new one")
        self.source_playfield.add_ball(balls=balls_to_safe)
        return {'balls': balls - balls_to_safe}

    def _ball_drain_count_balls(self, balls, **kwargs):
        del kwargs
        self.machine.events.post("multiball_{}_ball_lost".format(self.name))
        '''event: multiball_(name)_lost_ball
        desc: The multiball called (name) has lost a ball after ball save expired.
        '''

        if not self.machine.game or self.machine.game.balls_in_play - balls < 1:
            self.balls_added_live = 0
            self.balls_live_target = 0
            self.machine.events.remove_handler(self._ball_drain_count_balls)
            self.machine.events.post("multiball_{}_ended".format(self.name))
            '''event: multiball_(name)_ended
            desc: The multiball called (name) has just ended.
            '''
            self.debug_log("Ball drained. MB ended.")

    @event_handler(5)
    def event_stop(self, **kwargs):
        """Event handler for stop event."""
        del kwargs
        self.stop()

    def stop(self):
        """Stop shoot again."""
        self.debug_log("Stopping shoot again of multiball")
        self.shoot_again = False

        # disable shoot again
        self.machine.events.remove_handler(self._ball_drain_shoot_again)

        self.machine.events.post("multiball_" + self.name + "_shoot_again_ended")
        '''event: multiball_(name)_shoot_again_ended
        desc: Shoot again for multiball (name) has ended.
        '''

        # add handler for ball_drain until self.balls_ejected are drained
        self.machine.events.remove_handler(self._ball_drain_count_balls)
        self.machine.events.add_handler('ball_drain', self._ball_drain_count_balls)

    @event_handler(8)
    def event_add_a_ball(self, **kwargs):
        """Event handler for add_a_ball event."""
        del kwargs
        self.add_a_ball()

    def add_a_ball(self):
        """Add a ball if multiball has started."""
        if self.balls_live_target > 0:
            self.debug_log("Adding a ball.")
            self.balls_live_target += 1
            self.balls_added_live += 1
            self.machine.game.balls_in_play += 1
            self.source_playfield.add_ball(balls=1)

    @event_handler(9)
    def event_start_or_add_a_ball(self, **kwargs):
        """Event handler for start_or_add_a_ball event."""
        del kwargs
        self.start_or_add_a_ball()

    def start_or_add_a_ball(self):
        """Start multiball or add a ball if multiball has started."""
        if self.balls_live_target > 0:
            self.add_a_ball()
        else:
            self.start()

    @event_handler(2)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset the multiball and disable it."""
        self.disable()
        self.shoot_again = False
        self.balls_added_live = 0



"""Returns all annotated events by class."""
import ast
import os

import re
from collections import namedtuple

from typing import List

EventReference = namedtuple("EventReference", ["event_name", "file_name", "config_section", "class_label",
                                               "desc", "args", "config_attribute"])


class EventReferenceParser:

    """Parser to find all annotated events in MPF."""

    def _parse_file(self, file_name) -> List[EventReference]:
        """Parse one file and return all events from that file."""
        try:
            with open(file_name) as f:
                my_ast = ast.parse(f.read())
        except Exception:
            raise AssertionError("Error while parsing {}".format(file_name))

        event_list = []

        for x in ast.walk(my_ast):
            if isinstance(x, ast.ClassDef):
                class_label = None
                config_section = None
                for statement in x.body:
                    if isinstance(statement, ast.Assign) and len(statement.targets) == 1 and \
                       isinstance(statement.targets[0], ast.Name) and isinstance(statement.value, ast.Str):
                        if statement.targets[0].id == "class_label":
                            class_label = str(statement.value.s)
                        elif statement.targets[0].id == "config_section":
                            config_section = str(statement.value.s)

                for y in ast.walk(x):
                    if not (isinstance(y, ast.Str) and (y.s.strip().lower().startswith('event:'))):
                        continue

                    event_dict = self._parse_string(y)
                    if not event_dict:
                        continue

                    event_list.append(EventReference(
                        event_name=event_dict["event"], file_name=file_name,
                        config_section=config_section if not event_dict["config_section"] else
                        event_dict["config_section"],
                        class_label=class_label if not event_dict["class_label"] else event_dict["class_label"],
                        desc=event_dict["desc"],
                        args=self._parse_args(event_dict["args"]),
                        config_attribute=event_dict["config_attribute"]))

        return event_list

    def _parse_string(self, string):
        string = '\n'.join(' '.join(line.split())
                           for line in string.s.split('\n'))

        string = string.replace('Event:', 'event:')
        string = string.replace('Desc:', 'desc:')

        try:
            string = string.replace('Args:', 'args:')
        except ValueError:
            pass

        final_dict = self._string_to_args_dict(string, ['event', 'desc',
                                                        'args', 'config_attribute',
                                                        'config_section', 'class_label'])

        if not final_dict['desc']:
            # not an events docstring
            return None

        return final_dict

    def _parse_args(self, args_string):
        if args_string is None:
            return {}

        args = list()
        for x in re.findall(r'\b(\w*)\b(?=:)', args_string):
            if x:
                args.append(x)

        args_dict = self._string_to_args_dict(args_string, args)

        return args_dict

    @staticmethod
    def _string_to_args_dict(string, args):
        index_starts = list()
        for arg in args:
            try:
                index_starts.append(string.index(arg + ':'))
            except ValueError:
                pass

        index_starts.sort()
        sliced_list = list()
        for x, start in enumerate(index_starts):
            try:
                sliced_list.append(string[start:index_starts[
                    x + 1]])
            except IndexError:
                sliced_list.append(string[start:])

        final_dict = dict()

        for entry in sliced_list:
            split_entry = entry.split(':', 1)
            final_dict[split_entry[0].strip()] = split_entry[1].strip()

        for arg in args:
            if arg not in final_dict:
                final_dict[arg] = None

        return final_dict

    def get_events_from_path(self, paths) -> List[EventReference]:
        """Parse paths recursively and return all events."""
        events = []
        for path in paths:
            for root, _, files in os.walk(path):
                for file in [x for x in files if x.endswith('.py')]:
                    events.extend(self._parse_file(os.path.join(root, file)))

        return events



"""Contains the parent class for DEPRECATED Scriptlets.

This is deprecated and will be removed in config_version 6 with MPF 0.60.
Use custom code instead.
"""
from mpf.core.delays import DelayManager
from mpf.core.logging import LogMixin


class Scriptlet(LogMixin):

    """Baseclass for DEPRECATED scriptlets which are simple scripts in a machine.

    This is deprecated and will be removed in config_version 6 with MPF 0.60.
    Use custom code instead.
    """

    def __init__(self, machine, name):
        """Initialise scriptlet."""
        super().__init__()
        self.machine = machine
        self.name = name

        self.configure_logging('Scriptlet.' + name, 'basic', 'full')
        self.delay = DelayManager(self.machine)
        self.on_load()

    def __repr__(self):
        """Return string representation."""
        return '<Scriptlet.{}>'.format(self.name)

    def on_load(self):
        """Automatically called when this Scriptlet loads.

        It's the intention that the Scriptlet writer will overwrite this method
        in the Scriptlet.
        """



"""Mypinballs hardware platform."""
import re

from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplayPlatformInterface, FlashingType

from mpf.core.platform import SegmentDisplayPlatform


class MyPinballsSegmentDisplay(SegmentDisplayPlatformInterface):

    """A physical display on the mypinballs controller."""

    def __init__(self, number, platform) -> None:
        """Initialise segment display."""
        super().__init__(number)
        self.platform = platform        # type: MyPinballsHardwarePlatform

        # clear the display
        self.set_text("", FlashingType.NO_FLASH)

    def set_text(self, text: str, flashing: FlashingType):
        """Set digits to display."""
        if not text:
            # blank display
            cmd = b'3:' + bytes([ord(str(self.number))]) + b'\n'
        else:
            # remove any non-numbers and spaces
            text = re.sub(r'[^0-9 ]', "", text)

            # special char for spaces
            text = text.replace(" ", "?")
            # set text
            if flashing == FlashingType.FLASH_ALL:
                cmd = b'2:'
            elif flashing == FlashingType.FLASH_MATCH:
                cmd = b'4:'
            else:
                cmd = b'1:'
            cmd += bytes([ord(str(self.number))]) + b':' + text.encode() + b'\n'
        self.platform.send_cmd(cmd)


class MyPinballsHardwarePlatform(SegmentDisplayPlatform):

    """Hardware platform for MyPinballs 7-segment controller."""

    def __init__(self, machine):
        """Initialise mypinballs hardware."""
        super().__init__(machine)

        self._writer = None
        self._reader = None
        self.config = self.machine.config_validator.validate_config("mypinballs", self.machine.config['mypinballs'])
        self._configure_device_logging_and_debug("MyPinballs", self.config)

    async def initialize(self):
        """Initialise hardware."""
        # connect to serial
        connector = self.machine.clock.open_serial_connection(
            url=self.config['port'], baudrate=self.config['baud'])
        self._reader, self._writer = await connector

        # send a newline to end any previous command in the queue
        # this caused problems. disable it for now
        # self.send_cmd(b'\n')

    def stop(self):
        """Stop platform."""
        if self._writer:
            self._writer.close()

    def send_cmd(self, cmd: bytes):
        """Send a byte command."""
        if self.config['debug']:
            self.log.debug("Sending cmd: %s", cmd)
        self._writer.write(cmd)

    async def configure_segment_display(self, number: str, platform_settings) -> "SegmentDisplayPlatformInterface":
        """Configure display."""
        del platform_settings
        number_int = int(number)
        if 1 > number_int > 6:
            raise AssertionError("Number {} invalid for mypinballs display. 1-6 are valid.".format(number))

        return MyPinballsSegmentDisplay(number_int, self)



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestSystem11Trough(MpfTestCase):

    def get_config_file(self):
        return 'test_system_11_trough.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_boot_with_ball_in_drain_empty_trough(self):
        # MPF starts with a ball in the outhole (drain device). It should be
        # ejected into the trough and stay there.
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.machine.switch_controller.process_switch("outhole", 1)
        self.advance_time_and_run(.6)

        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)

        self.machine.switch_controller.process_switch("outhole", 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("trough3", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(1, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

    def test_boot_with_balls_in_drain_and_trough(self):
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.machine.switch_controller.process_switch("outhole", 1)
        self.machine.switch_controller.process_switch("trough1", 1)
        self.machine.switch_controller.process_switch("trough2", 1)
        self.advance_time_and_run(.6)

        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)

        self.machine.switch_controller.process_switch("outhole", 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("trough3", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

    def test_add_ball_to_pf(self):
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.machine.switch_controller.process_switch("trough1", 1)
        self.machine.switch_controller.process_switch("trough2", 1)
        self.machine.switch_controller.process_switch("trough3", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)
        self.machine.switch_controller.process_switch("trough1", 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.process_switch("plunger", 1)

        self.advance_time_and_run(1)

        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('ejecting', self.machine.ball_devices["plunger"]._state)
        # self.assertEquals(1,
        #                   self.machine.ball_devices["playfield"].num_balls_requested)

        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(1)

    def test_single_ball_drain_and_eject(self):
        # tests that when a ball drains into the outhole and is in the process
        # of being ejected to the trough, MPF is able to also request a ball
        # to the plunger even though the ball hasn't made it into the trough
        # yet

        self.machine.switch_controller.process_switch("trough1", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(1, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)

        # start a game, ball goes to plunger
        self.machine.switch_controller.process_switch("start", 1)
        self.machine.switch_controller.process_switch("start", 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("trough1", 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("plunger", 1)
        self.advance_time_and_run(2)

        self.assertIsNotNone(self.machine.game)

        # plunge to playfield
        self.machine.switch_controller.process_switch("plunger", 0)
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(2)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)

        # drain
        self.machine.switch_controller.process_switch("outhole", 1)
        self.advance_time_and_run(.6)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(2, self.machine.game.player.ball)

        self.machine.switch_controller.process_switch("outhole", 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("trough1", 1)


class TestSystem11TroughStartup(MpfTestCase):

    def get_config_file(self):
        return 'test_system_11_trough_startup.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def get_platform(self):
        return "smart_virtual"

    def test_start_with_four_balls(self):
#        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)
        self.assertEqual(1, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["outhole"].available_balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(4, self.machine.ball_devices["trough"].available_balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)

        self.hit_and_release_switch("start")
        self.advance_time_and_run(20)
        self.release_switch_and_run("plunger", 10)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)



"""Show config player."""
from mpf.core.placeholder_manager import ConditionalEvent

from mpf.config_players.device_config_player import DeviceConfigPlayer

RESERVED_KEYS = ["show", "priority", "speed", "block_queue", "start_step", "loops", "sync_ms", "manual_advance",
                 "key", "show_tokens", "events_when_played", "events_when_stopped", "events_when_looped",
                 "events_when_paused", "events_when_resumed", "events_when_advanced",
                 "events_when_stepped_back", "events_when_updated", "events_when_completed"]


class ShowPlayer(DeviceConfigPlayer):

    """Plays, starts, stops, pauses, resumes or advances shows based on config."""

    config_file_section = 'show_player'
    show_section = 'shows'
    allow_placeholders_in_keys = True

    __slots__ = ["_actions"]

    def __init__(self, machine):
        """Initialise show player."""
        super().__init__(machine)
        self._actions = {
            'play': self._play,
            'stop': self._stop,
            'pause': self._pause,
            'resume': self._resume,
            'advance': self._advance,
            'step_back': self._step_back,
            'update': self._update,
            'queue': self._queue
        }

    # pylint: disable-msg=too-many-arguments
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Play, start, stop, pause, resume or advance show based on config."""
        # make sure all shows play in sync
        queue = kwargs.get("queue", None)
        start_time = kwargs.get("start_time", None)
        show_tokens = kwargs.get("show_tokens", kwargs)
        if not start_time:
            start_time = self.machine.clock.get_time()
        for show, show_settings in settings.items():
            # Look for a conditional event in the show name
            if show.condition and not show.condition.evaluate(kwargs):
                continue

            if 'hold' in show_settings and show_settings['hold'] is not None:
                raise AssertionError(
                    "Setting 'hold' is no longer supported for shows. Use duration -1 in your show.")
            if priority:
                show_settings = dict(show_settings)
                try:
                    show_settings['priority'] += priority
                except KeyError:
                    show_settings['priority'] = priority

            self._update_show(show.name, show_settings, context, queue, start_time, show_tokens)

    def _expand_device(self, device):
        # parse conditionals
        devices = super()._expand_device(device)
        for index, device_entry in enumerate(devices):
            if not isinstance(device_entry, ConditionalEvent):
                devices[index] = self.machine.placeholder_manager.parse_conditional_template(device_entry)
        return devices

    def _expand_device_config(self, device_settings):
        """Validate show_tokens."""
        for key in RESERVED_KEYS:
            if key in device_settings["show_tokens"]:
                self.raise_config_error("Key {} is not allowed in show_tokens of your show_player because it is also "
                                        "an option in show_player. Did you indent that option too far?".format(key), 1)
        return device_settings

    def handle_subscription_change(self, value, settings, priority, context, key):
        """Handle subscriptions."""
        instance_dict = self._get_instance_dict(context)
        for show, show_settings in settings.items():
            show_settings = dict(show_settings)

            show_key = show_settings["key"] if 'key' in show_settings and show_settings['key'] else key

            if show_settings['action'] != 'play':
                raise AssertionError("Can only use action play with subscriptions.")

            if value:
                self._play(show_key, instance_dict, show.name, show_settings, False, None, {})
            else:
                self._stop(show_key, instance_dict, show.name, show_settings, False, None, {})

    # pylint: disable-msg=too-many-arguments
    def _play(self, key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        stop_callback = None
        if show_settings['block_queue']:
            if not queue:
                raise AssertionError("block_queue can only be used with a queue event.")
            queue.wait()
            stop_callback = queue.clear
        start_step = show_settings['start_step'].evaluate(placeholder_args)
        start_running = show_settings['start_running'].evaluate(placeholder_args)
        show_tokens = {k: v.evaluate(placeholder_args) for k, v in show_settings['show_tokens'].items()}

        show_config = self.machine.show_controller.create_show_config(
            show, show_settings['priority'], show_settings['speed'], show_settings['loops'], show_settings['sync_ms'],
            show_settings['manual_advance'], show_tokens, show_settings['events_when_played'],
            show_settings['events_when_stopped'], show_settings['events_when_looped'],
            show_settings['events_when_paused'], show_settings['events_when_resumed'],
            show_settings['events_when_advanced'], show_settings['events_when_stepped_back'],
            show_settings['events_when_updated'], show_settings['events_when_completed'])

        previous_show = instance_dict.get(key, None)

        instance_dict[key] = self.machine.show_controller.replace_or_advance_show(previous_show, show_config,
                                                                                  start_step, start_time,
                                                                                  start_running, stop_callback)

    # pylint: disable-msg=too-many-arguments
    def _queue(self, key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del queue
        del instance_dict
        del start_time
        del key
        if show_settings['block_queue']:
            raise AssertionError("Cannot use queue with block_queue.")

        start_step = show_settings['start_step'].evaluate(placeholder_args)
        show_tokens = {k: v.evaluate(placeholder_args) for k, v in show_settings['show_tokens'].items()}

        show_config = self.machine.show_controller.create_show_config(
            show, show_settings['priority'], show_settings['speed'], show_settings['loops'], show_settings['sync_ms'],
            show_settings['manual_advance'], show_tokens, show_settings['events_when_played'],
            show_settings['events_when_stopped'], show_settings['events_when_looped'],
            show_settings['events_when_paused'], show_settings['events_when_resumed'],
            show_settings['events_when_advanced'], show_settings['events_when_stepped_back'],
            show_settings['events_when_updated'], show_settings['events_when_completed'])

        show_settings["show_queue"].enqueue_show(show_config, start_step)

    @staticmethod
    def _stop(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del show_settings
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].stop()
            del instance_dict[key]

    @staticmethod
    def _pause(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del show_settings
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].pause()

    @staticmethod
    def _resume(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del show_settings
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].resume()

    @staticmethod
    def _advance(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del show_settings
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].advance()

    @staticmethod
    def _step_back(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del show_settings
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].step_back()

    @staticmethod
    def _update(key, instance_dict, show, show_settings, queue, start_time, placeholder_args):
        del show
        del queue
        del start_time
        del placeholder_args
        if key in instance_dict:
            instance_dict[key].update(
                speed=show_settings.get('speed'),
                manual_advance=show_settings.get('manual_advance')
            )

    # pylint: disable-msg=too-many-arguments
    def _update_show(self, show, show_settings, context, queue, start_time, placeholder_args):
        instance_dict = self._get_instance_dict(context)
        if 'key' in show_settings and show_settings['key']:
            key = show_settings['key']
        else:
            key = show

        try:
            action = self._actions[show_settings['action']]
        except KeyError:
            raise AssertionError("Invalid action {} in show_player {}".format(
                show_settings['action'], key))

        if 'show' in show_settings and show_settings['show']:
            show_name = show_settings['show']
        else:
            show_name = show

        action(key, instance_dict, show_name, show_settings, queue, start_time, placeholder_args)

    def clear_context(self, context):
        """Stop running shows from context."""
        for show in self._get_instance_dict(context).values():
            show.stop()
        self._reset_instance_dict(context)

    def get_express_config(self, value):
        """Parse express config."""
        return {"action": value}



"""Special events for the docs."""


class SpecialEvents:

    """A class to document special events.

    This is a legacy thing until the docs parse mpfconfig.yaml to generate those or we find another way.
    """

    '''event: flipper_cancel

    desc: Posted when both flipper buttons are hit at the same time,
    useful as a "cancel" event for shows, the bonus mode, etc.

    Note that in order for this event to work, you have to add
    ``left_flipper`` as a tag to the switch for your left flipper,
    and ``right_flipper`` to your right flipper.

    See :doc:`/config/combo_switches` for details.
    '''

    '''event: flipper_cradle
    config_attribute: events_when_active

    desc: Posted when one of the flipper buttons has been active for 3
    seconds.

    Note that in order for this event to work, you have to add
    ``left_flipper`` as a tag to the switch for your left flipper,
    and ``right_flipper`` to your right flipper.

    See :doc:`/config/timed_switches` for details.
    '''

    '''event: flipper_cradle_release
    config_attribute: events_when_released

    desc: Posted when one of the flipper buttons that has previously
    been active for more than 3 seconds has been released.

    If the player pushes in one flipper button for more than 3 seconds,
    and then the second one and holds it in for more than 3 seconds,
    this event won't be posted until both buttons have been released.

    Note that in order for this event to work, you have to add
    ``left_flipper`` as a tag to the switch for your left flipper,
    and ``right_flipper`` to your right flipper.

    See :doc:`/config/timed_switches` for details.
    '''



"""Base classes and utilities for wiring."""
from typing import Dict, List, Tuple, TypeVar, Callable

wire_colors = {
    12: "YE",
    5: "RD",
    48: "BU",
    0: "BK",
    -2: "WH",
    1: "WH",
    -1: "PK"
}

PinSpecs = List[Tuple[str, int]]


class Board:

    """A board with connectors. This is not an element in wireviz, but useful to track."""

    def __init__(self, board_type: str):
        """Create a board with the given type."""
        # Note: board_type is used because "type" is reserved.
        self.connectors: Dict[str, Connector] = dict()
        self.board_type = board_type
        self.ordinal = 0

    def add_connector(self, name: str, pinspecs: PinSpecs):
        """Add a new connector to this board, with the listed set of pins."""
        new_connector = Connector(name, pinspecs, self)
        self.connectors[name] = new_connector

    def __repr__(self):
        """Output simple string representation of board for debugging."""
        s = "Board " + self.board_type + ":\n"
        for c in self.connectors.values():
            s = s + str(c) + "\n"
        return s

    def __getitem__(self, key):
        """Allow indexing to be used to access connectors dictionary."""
        return self.connectors[key]

    def dump(self) -> dict:
        """Output this board in a form suitable for passing to yaml output."""
        # Boards are not separate entities in wireviz, so just combine all connectors
        o = dict()
        for c in self.connectors.values():
            o.update(c.dump())
        return o


class Connector:

    """A single connector on a board."""

    def __init__(self, name: str, pinspecs: PinSpecs, board: Board):
        """Create a connector with the given name and pin specifications on the given board.

        Easier to use board.add_connector().
        """
        self.name = name
        self.pins: List[Pin] = []
        self.board = board
        for od, (pin_name, vc) in enumerate(pinspecs):
            self.pins.append(Pin(pin_name, vc, self, od))

    def __repr__(self):
        """Output simple string representation of connector for debugging."""
        s = self.name + ":\n"
        for pi, p in enumerate(self.pins):
            s = s + str(pi) + ":" + str(p) + "\n"
        return s

    def __getitem__(self, key):
        """Allow indexing to be used to access pins."""
        return self.pins[key]

    def get_display_name(self) -> str:
        """Get the name of this connector to be shown in wireviz."""
        if self.name != "":
            return self.board.board_type + " " + str(self.board.ordinal) + " " + self.name
        return self.board.board_type

    def dump(self) -> dict:
        """Output this connector in a form suitable for passing ot yaml output."""
        display_name = self.get_display_name()
        pin_labels: List[str] = []
        for p in self.pins:
            pin_labels.append(p.name)

        return {display_name: {"pincount": len(self.pins), "pinlabels": pin_labels}}


class Pin:

    """A pin on a connector.

    In the case of a connector such as USB or RJ45, where the pin assignments
    aren't relevant, may represent the whole connector.
    """

    def __init__(self, name: str, vclass: int, connector: Connector, ordinal: int):
        """Initialize a pin with given name and voltage class, on the given board.

        This should probably be done through addConnector() since the pin's position is needed.
        """
        self.name = name
        self.vclass = vclass
        self.connector = connector
        self.ordinal = ordinal

    def __repr__(self):
        """Return simple string debug output for pin."""
        return self.name + " " + str(self.vclass)


class Wire:

    """Represents a wire connecting two pins.

    In reality electrical wires don't have a 'direction', but wireviz puts sources on the
    left and destinations on the right, so we track them.
    """

    def __init__(self, src: Pin, dest: Pin):
        """Initialize a wire connecting the two specified pins.

        This isn't a lot of use outside of a System, so you probably want to use System.connect().
        """
        if src.vclass != dest.vclass:
            print("Connecting pins of different vclasses!")
            print(src, dest)
        self.src = src
        self.dest = dest

    def __eq__(self, other):
        """Wires that connect the same pins are equal, regardless of order."""
        if other is not Wire:
            return False
        return self.connects_pins(other.src, other.dest)

    def connects_pins(self, a: Pin, b: Pin) -> bool:
        """Return if the wire connects these two pins (in either direction)."""
        if self.src == a and self.dest == b:
            return True
        if self.src == b and self.dest == a:
            return True
        return False

    def connects_connectors(self, a: Connector, b: Connector) -> bool:
        """Return if the wire connects these two connectors (in either direction and any pin)."""
        if self.src.connector == a and self.dest.connector == b:
            return True
        if self.src.connector == b and self.dest.connector == a:
            return True
        return False


class SerialLED(Board):

    """A basic serial LED."""

    def __init__(self, name):
        """Initialize the LED with the WS2812 pinout."""
        super().__init__("WS2812 " + name)
        self.add_connector("", [
            ("DOUT", 5),
            ("DIN", 5),
            ("VCC", 5),
            ("NC", -1),
            ("VDD", 5),
            ("GND", 0)])


class Switch(Board):

    """A basic switch."""

    def __init__(self, name):
        """Initialize the switch with + and - sides."""
        super().__init__("SW " + name)
        self.add_connector("", [
            ("+", 5),
            ("-", 5)])


class Coil(Board):

    """A basic coil/driver."""

    def __init__(self, name):
        """Initialize the coil with + and - sides.

        TODO: Flipper coils may have multiple negative sides for hold and flip.
        """
        super().__init__("DR " + name)
        self.add_connector("", [
            ("+", 48),
            ("-", 48)])


# pylint: disable=invalid-name
T = TypeVar("T")


class System:

    """A system of wires, boards and connectors."""

    def __init__(self):
        """Initialize an empty system."""
        self.boards: List[Board] = []
        self.wires: List[Wire] = []

    def add_board(self, board: Board):
        """Add a board to the system."""
        # Set ordinal based on number of existing boards of same type.
        o = sum([1 for b in self.boards if b.board_type == board.board_type])
        board.ordinal = o
        self.boards.append(board)

    def connect(self, src: Pin, dest: Pin):
        """Add a wire between two pins."""
        assert src.connector.board in self.boards
        assert dest.connector.board in self.boards
        nw = Wire(src, dest)
        if nw not in self.wires:
            self.wires.append(nw)

    def daisy_chain_list(self, items: List[T], get_in: Callable[[T], Pin], get_out: Callable[[T], Pin]):
        """Daisy chains connections between arbitrary items that can calculate pins.

        :param items The list of items, of any type.
        :param get_in Function to apply to an item to get the input pin.
        :param get_out Function to apply to an item to get the output pin.
        """
        if len(items) < 2:
            return
        for index in range(1, len(items)):
            self.connect(get_out(items[index - 1]), get_in(items[index]))

# pylint: disable=too-many-arguments
    def daisy_chain_dict(self, items: Dict[int, T], get_in: Callable[[T], Pin], get_out: Callable[[T], Pin],
                         start: int, ladder: bool) -> bool:
        """Like daisy_chain_list but takes a dict and checks it for sequentiality in the process of daisy chaining.

        Used for directly daisy chaining elements with specified numbers.
        :param items The dictionary from numbers to items, of any type.
        :param get_in Function to apply to an item to get the input pin.
        :param get_out Function to apply to an item to get the output pin.
        :param start Number to start accessing the dictionary from.
        :param ladder If true, alternate chain connections are flipped to create a vertical ladder in wireviz.
        :return True if all items in the dictionary were sequentially numbered. If not, the chain stops
                     at the first gap.
        """
        if len(items) < 2:
            return True
        if start not in items:
            return False
        even = False
        for index in range(start + 1, start + len(items)):
            if index not in items:
                return False
            if even or not ladder:
                self.connect(get_out(items[index - 1]), get_in(items[index]))
            else:
                self.connect(get_in(items[index]), get_out(items[index - 1]))
            even = not even
        return True

    # pylint: disable=too-many-locals
    def dump(self) -> dict:
        """Output this system in a format suitable for yaml output."""
        # Output all connectors
        connectors_dict = dict()       # Connectors dictionary for YAML
        for board in self.boards:
            connectors_dict.update(board.dump())

        # Calculate list of all pairs of connectors (NB not pins) connected by wires
        pairs: List[Tuple[Connector, Connector]] = []
        for wire in self.wires:
            if (wire.src.connector, wire.dest.connector) not in pairs:
                if (wire.dest.connector, wire.src.connector) not in pairs:
                    pairs.append((wire.src.connector, wire.dest.connector))

        wire_dict = dict()              # Wires dictionary for YAML
        connection_list: list = []      # Connections list for YAML
        wire_ordinal = 0                # Serial number for next wire

        for (srcc, destc) in pairs:
            # Find all wires that connect each pair
            wires_this_pair: List[Wire] = [wire for wire in self.wires if wire.connects_connectors(srcc, destc)]

            # Connect them into a single multi-thread "wire" for wireviz
            # Calculate name for the wire based on current ordinal
            compound_wire_name = "W" + str(wire_ordinal)
            wire_ordinal += 1

            # Wireviz requires three lists per wire set, matched in order: source pins, wire thread numbers,
            # destination pins
            src_pin_list = []
            wire_list = []
            dest_pin_list = []

            # Wire colours for wire specifier
            color_list = []

            for x, awire in enumerate(wires_this_pair):
                src_pin_list.append(awire.src.ordinal + 1)
                wire_list.append(x + 1)
                dest_pin_list.append(awire.dest.ordinal + 1)
                color_list.append(wire_colors[awire.src.vclass])

            # Weird wireviz format for the cables block: a list of single entry dictionaries of lists
            connection_dict = [{srcc.get_display_name(): src_pin_list}, {compound_wire_name: wire_list},
                               {destc.get_display_name(): dest_pin_list}]
            connection_list.append(connection_dict)

            # Add entry to wires block
            wire_dict.update({compound_wire_name: {"wirecount": len(wires_this_pair), "colors": color_list}})

        return {"connectors": connectors_dict, "cables": wire_dict, "connections": connection_list}



import logging
import os
import re
import tempfile

import shutil
import shlex

from mpf.tests.MpfGameTestCase import MpfGameTestCase

from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip

from mpf.file_interfaces.yaml_interface import YamlInterface

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfMachineTestCase import MockConfigPlayers


class MpfDocTestCaseBase(MockConfigPlayers, MpfGameTestCase):
    def __init__(self, config_string, base_dir = None, simulation=True, methodName='test_config_parsing'):
        super().__init__(methodName)
        self._config_string = config_string
        self._base_dir = base_dir
        self._simulation = simulation
        self.log = logging.getLogger("TEST")

    def setUp(self):
        machine_config, extra_configs, mode_configs, show_configs, assets, code, self.tests = \
            self.prepare_config(self._config_string)
        self.config_dir = tempfile.mkdtemp()
        # cleanup at the end
        self.addCleanup(self._delete_tmp_dir, self.config_dir)

        # create machine config
        os.mkdir(os.path.join(self.config_dir, "config"))
        self.verify_yaml(machine_config, "config/config.yaml")
        with open(os.path.join(self.config_dir, "config/", "config.yaml"), "w") as f:
            f.write(machine_config)

        # create shows
        os.mkdir(os.path.join(self.config_dir, "shows"))
        for show_name, show_config in show_configs.items():
            self.verify_yaml(show_config, "shows/" + show_name + ".yaml", show_file=True)
            with open(os.path.join(self.config_dir, "shows", show_name + ".yaml"), "w") as f:
                f.write(show_config)

        # create modes
        os.mkdir(os.path.join(self.config_dir, "modes"))
        for mode_name, mode_config in mode_configs.items():
            self.verify_yaml(mode_config, "modes/" + mode_name + "/config/" + mode_name + ".yaml")
            os.mkdir(os.path.join(self.config_dir, "modes", mode_name))
            os.mkdir(os.path.join(self.config_dir, "modes", mode_name, "config"))
            with open(os.path.join(self.config_dir, "modes", mode_name, "config", mode_name + ".yaml"), "w") as f:
                f.write(mode_config)

        # create code files
        for file_path, file_content in code:
            os.mkdir(os.path.join(self.config_dir, *file_path[:-1]))
            with open(os.path.join(self.config_dir, *file_path), "w") as f:
                f.write(file_content)

        # create extra configs
        for config_name, extra_config in extra_configs.items():
            self.verify_yaml(extra_config, config_name + ".yaml")
            with open(os.path.join(self.config_dir, config_name + ".yaml"), "w") as f:
                f.write(extra_config)

        # link assets
        for asset_path, asset_source in assets.items():
            if not self._base_dir:
                raise AssertionError("You need to set base_dir to use assets.")
            path_elements = asset_path.split("/")
            source_elements = asset_source.split("/")
            full_source_path = os.path.join(self._base_dir, *source_elements)
            try:
                os.mkdir(os.path.join(self.config_dir, *path_elements[:-1]))
            except FileExistsError:
                pass
            self.get_absolute_machine_path()
            if not os.path.isfile(full_source_path):
                raise AssertionError('Could not find asset "{}" on disk'.format(full_source_path))
            os.symlink(full_source_path, os.path.join(self.config_dir, *path_elements))

        super().setUp()

    @staticmethod
    def unidiff_output(expected, actual):
        """
        Helper function. Returns a string containing the unified diff of two multiline strings.
        """

        import difflib
        expected = expected.splitlines(1)
        actual = actual.splitlines(1)

        diff = difflib.unified_diff(expected, actual)

        return ''.join(diff)

    def verify_yaml(self, yaml_string, config_name, show_file=False):
        formatted_yaml = YamlRoundtrip.reformat_yaml(yaml_string, show_file=show_file)
        if formatted_yaml == "null\n...\n":
            # special case: empty config
            return
        if yaml_string.strip() != formatted_yaml.strip():
            #self.maxDiff = None
            #self.assertEqual(yaml_string, formatted_yaml)
            diff = self.unidiff_output(yaml_string.strip() + "\n", formatted_yaml.strip() + "\n")
            self.fail("Config {} unformatted. Diff:\n{}".format(
              config_name, diff))

    def get_enable_plugins(self):
        return True

    def get_options(self):
        options = super().get_options()
        # no cache since we are in a tmp folder anyway
        options['no_load_cache'] = True,
        options['create_config_cache'] = False
        return options

    @staticmethod
    def prepare_config(config_string, fixup_config=True):
        # inline invisible comments from documentation
        if fixup_config:
            config_string = re.sub(r'^#! ([^\n]+)', '\\1', config_string, flags=re.MULTILINE)

        # first find sections
        configs = re.split(r'^##! (config: [^\n]+|test|show: \w+|mode: \w+|asset: [^\n]+|code: [^\n]+)\n', config_string,
                           flags=re.MULTILINE)
        machine_config = configs.pop(0)

        # add config_version if missing
        if fixup_config and not machine_config.startswith("#config_version=5"):
            machine_config = "#config_version=5\n" + machine_config

        modes = {}
        shows = {}
        tests = []
        assets = {}
        extra_configs = {}
        game_code = []
        while configs:
            section_type = configs.pop(0)
            section_config = configs.pop(0)
            if section_type.startswith("test"):
                # unit test
                tests = section_config.splitlines()
            elif section_type.startswith("config:"):
                # extra config file
                _, file_name = section_type.split(": ", 2)
                if fixup_config and not section_config.startswith("#config_version=5"):
                    section_config = "#config_version=5\n" + section_config
                extra_configs[file_name] = section_config
            elif section_type.startswith("mode:"):
                # normal mode
                _, mode_name = section_type.split(": ", 2)
                if fixup_config and not section_config.startswith("#config_version=5"):
                    section_config = "#config_version=5\n" + section_config
                modes[mode_name] = section_config
            elif section_type.startswith("show:"):
                # show
                _, show_name = section_type.split(": ", 2)
                if fixup_config and not section_config.startswith("#show_version=5"):
                    section_config = "#show_version=5\n" + section_config
                shows[show_name] = section_config
            elif section_type.startswith("asset:"):
                # asset
                _, asset_desc = section_type.split(": ", 2)
                asset_path, asset_source = asset_desc.split("=", 2)
                assets[asset_path] = asset_source
            elif section_type.startswith("code:"):
                # asset
                _, file_location = section_type.split(": ", 2)
                game_code.append((file_location.split("/"), section_config))
            else:
                raise AssertionError("Invalid section: {}".format(section_type))

        # load all modes
        if fixup_config and modes and "modes:" not in machine_config:
            machine_config += "\nmodes:\n"
            for mode in modes.keys():
                machine_config += "  - " + mode + "\n"

        return machine_config, extra_configs, modes, shows, assets, game_code, tests

    def get_config_file(self):
        return "config.yaml"

    def get_platform(self):
        if self._simulation:
            return "smart_virtual"
        else:
            return "virtual"

    def get_machine_path(self):
        return self.config_dir

    def _delete_tmp_dir(self, config_dir):
        shutil.rmtree(config_dir)

    def test_config_parsing(self):
        line_no = 0
        for line in self.tests:
            line_no += 1
            if not line or line.startswith("#"):
                continue

            parts = shlex.split(line)
            command = parts.pop(0)
            method = getattr(self, "command_" + command)
            if not method:
                raise AssertionError("Unknown command {} in line {}".format(command, line_no))
            self.log.info("%s", line)
            try:
                method(*parts)
            except AssertionError as e:
                raise AssertionError("Error in assertion {} (Num: {}) of your tests.".format(command, line_no)) from e

    def command_add_player(self):
        self.add_player()

    def command_stop_game(self, stop_time=1):
        self.stop_game(float(stop_time))

    def command_start_two_player_game(self):
        self.start_two_player_game()

    def command_drain_all_balls(self):
        self.drain_all_balls()

    def command_drain_one_ball(self):
        self.drain_one_ball()

    def command_add_ball_to_device(self, device_name):
        if device_name not in self.machine.ball_devices:
            raise AssertionError("Invalid ball device {}".format(device_name))
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices[device_name])

    def command_start_mode(self, mode):
        self.machine.modes[mode].start()
        self.machine_run()
        self.assertModeRunning(mode)

    def command_stop_mode(self, mode):
        self.machine.modes[mode].stop()
        self.machine_run()
        self.assertModeNotRunning(mode)

    def command_assert_mode_running(self, mode):
        self.assertModeRunning(mode)

    def command_assert_mode_not_running(self, mode):
        self.assertModeNotRunning(mode)

    def command_post(self, event_name, *args):
        kwargs = {}
        for arg in args:
            key, value = arg.split("=", 2)
            kwargs[key] = value
        self.post_event_with_params(event_name, **kwargs)

    def command_hit_and_release_switch(self, switch_name):
        self.hit_and_release_switch(switch_name)

    def command_hit_and_release_switches_simultaneously(self, switch_name1, switch_name2):
        self.hit_and_release_switches_simultaneously([switch_name1, switch_name2])

    def command_hit_switch(self, switch_name):
        self.hit_switch_and_run(switch_name, 0)

    def command_release_switch(self, switch_name):
        self.release_switch_and_run(switch_name, 0)

    def command_advance_time_and_run(self, delta):
        self.advance_time_and_run(float(delta))

    def command_assert_player_variable(self, value, player_var):
        if isinstance(self.machine.game.player[player_var], (int, float)):
            value = float(value)
        self.assertPlayerVarEqual(value, player_var=player_var)

    def command_assert_player_count(self, count):
        self.assertPlayerCount(int(count))

    def command_assert_machine_variable(self, value, name):
        if name in self.machine.variables.machine_vars and isinstance(
                self.machine.variables.machine_vars[name]["value"], (int, float)):
            value = float(value)
        self.assertMachineVarEqual(value, name)

    def command_assert_light_color(self, light, color):
        self.assertLightColor(light, color)

    def command_assert_light_flashing(self, light, color, secs=1, delta=.1):
        self.assertLightFlashing(light, color, float(secs), float(delta))

    def command_mock_event(self, name):
        self.mock_event(name)

    def command_assert_event_called(self, name, times=1):
        self.assertEventCalled(name, int(times))

    def command_assert_event_not_called(self, name):
        self.assertEventNotCalled(name)

    def command_assert_balls_on_playfield(self, balls):
        self.assertBallsOnPlayfield(int(balls))

    def command_assert_available_balls_on_playfield(self, balls):
        self.assertAvailableBallsOnPlayfield(int(balls))

    def command_assert_balls_in_play(self, balls):
        self.assertBallsInPlay(int(balls))

    def command_assert_str_condition(self, expected, condition):
        expected_str = str(expected)
        self.assertPlaceholderEvaluates(expected_str, condition)

    def command_assert_bool_condition(self, expected, condition):
        expected_bool = expected == "True"
        self.assertPlaceholderEvaluates(expected_bool, condition)

    def command_assert_int_condition(self, expected, condition):
        expected_int = int(expected)
        self.assertPlaceholderEvaluates(expected_int, condition)

    def command_assert_float_condition(self, expected, condition):
        expected_float = float(expected)
        self.assertPlaceholderEvaluates(expected_float, condition)


class MpfDocTestCaseNoFakeGame(MpfDocTestCaseBase):

    """Test case with real game."""

    def command_start_game(self):
        self.start_game()


class MpfDocTestCase(MpfDocTestCaseBase, MpfFakeGameTestCase):

    """Test case with fake game."""

    def command_start_game(self, num_balls_known=3):
        self.start_game(num_balls_known=int(num_balls_known))



"""A Power Supply Unit (PSU) in a pinball machine."""
from mpf.core.system_wide_device import SystemWideDevice


class PowerSupplyUnit(SystemWideDevice):

    """Represents a power supply in a pinball machine."""

    config_section = 'psus'
    collection = 'psus'
    class_label = 'psu'

    __slots__ = ["_busy_until"]

    def __init__(self, machine, name):
        """Initialise PSU."""
        super().__init__(machine, name)
        self._busy_until = None

    def get_wait_time_for_pulse(self, pulse_ms, max_wait_ms) -> int:
        """Return a wait time for a pulse or 0."""
        current_time = self.machine.clock.get_time()

        if self._busy_until and self._busy_until < current_time:
            # prevent negative times
            self._busy_until = None

        if not self._busy_until or not max_wait_ms:
            # if we are not busy. do pulse now
            self.notify_about_instant_pulse(pulse_ms)
            return 0

        if self._busy_until > current_time + (max_wait_ms / 1000.0) or max_wait_ms is None:
            # if we are busy for longer than possible. do pulse now
            self.notify_about_instant_pulse(pulse_ms)
            return 0

        # calculate wait time and return it
        wait_ms = (self._busy_until - current_time) * 1000
        self._busy_until += (pulse_ms + self.config['release_wait_ms']) / 1000.0
        return wait_ms

    def notify_about_instant_pulse(self, pulse_ms):
        """Notify PSU about pulse."""
        if self._busy_until:
            self._busy_until = max(
                self._busy_until,
                self.machine.clock.get_time() + (pulse_ms + self.config['release_wait_ms']) / 1000.0)
        else:
            self._busy_until = self.machine.clock.get_time() + ((pulse_ms + self.config['release_wait_ms']) / 1000.0)



from mpf.tests.MpfDocTestCase import MpfDocTestCase
from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase


class MpfIntegrationDocTestCase(MpfDocTestCase, MpfIntegrationTestCase, MpfSlideTestCase):

    def command_assert_slide_active(self, name):
        self.assertSlideActive(name)

    def command_assert_slide_not_active(self, name):
        self.assertSlideNotActive(name)

    def command_assert_text_in_slide(self, text, slide_name):
        self.assertTextInSlide(text, slide_name)

    def command_assert_text_not_in_slide(self, text, slide_name):
        self.assertTextNotInSlide(text, slide_name)

    def command_assert_slide_on_top(self, slide_name, target="default"):
        self.assertSlideOnTop(slide_name, target)

    def command_assert_text_on_top_slide(self, text, target="default"):
        self.assertTextOnTopSlide(text, target)

    def command_assert_text_not_on_top_slide(self, text, target="default"):
        self.assertTextNotOnTopSlide(text, target)